#!/usr/bin/perl -w
# $Id: dnsmon-hourly-cronjob.pl,v 1.6.2.1 2008/04/13 15:49:47 ruben Exp $
############################################################################
###    (C)opyright 2003 - 2008 RIPE NCC
###    This file is part of DNSMon
###
###    DNSMon is free software: you can redistribute it and/or modify
###    it under the terms of the GNU General Public License as published by
###    the Free Software Foundation, either version 3 of the License, or
###    (at your option) any later version.
###
###    DNSMon is distributed in the hope that it will be useful,
###    but WITHOUT ANY WARRANTY; without even the implied warranty of
###    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
###    GNU General Public License for more details.
###
###    You should have received a copy of the GNU General Public License
###    along with DNSMon.  If not, see <http://www.gnu.org/licenses/>.
############################################################################

eval 'exec /usr/bin/perl -w -S $0 ${1+"$@"}'
    if 0; # not running under some shell
use strict;
use warnings;
use Data::Dumper;
use POSIX;
use Socket;
use Socket6;

use Time::Piece;
use LockFile::Simple;
use Fcntl qw(LOCK_EX);
use Carp ();
use Log::Log4perl 'get_logger';
use Time::HiRes qw(gettimeofday tv_interval);
use Getopt::Std;

use RIPE::DNSMon::RRD;             # dns rrd routines

use constant GENPLOTS_DAILY_HOUR => 1;	# when to run daily task, in UTC
use constant TIME_TO_LIVE => 3600;	# number of seconds this task is expected to fit in
					# in other words, period of time we run it from cron

sub report;
sub domplot;
sub serplot;
sub monplot;
sub _die;
sub updaterrds(\@$);
sub waitforchild;
sub run_in_parallel (&@);

Log::Log4perl::init("$ENV{'HOME'}/.log4perl.conf");

$ENV{LANG} = "POSIX";

my $logger = get_logger('dnsmon.dnsmon-hourly-cronjob');

my $nrunning : shared =0;		# number of child processes running
my $nconcur=1;		# max number of simultaneous procs
my %child_t;	# track processing time per child
my %opt;

# Only enable on newer versions
$Getopt::Std::STANDARD_HELP_VERSION = 1 if $Getopt::Std::VERSION >= 1.05;

getopts('uphj:',\%opt);

my $doplots = 1;
my $doupdates = 1;

if (exists($opt{'u'}) && exists($opt{'p'})) {
	&usage;
} elsif (exists($opt{'u'})) {
	$doupdates = 1;
	$doplots = 0;
} elsif (exists($opt{'p'})) {
	$doupdates = 0;
	$doplots = 1;
}

&usage() if (exists($opt{'h'}));

my $osname =  ((uname())[0]);

if ($osname =~ /FreeBSD/) {
	$nconcur = qx(/sbin/sysctl -n hw.ncpu);
	chomp($nconcur);
}
elsif ($osname =~ /Linux/) {
	$nconcur = qx(cat /proc/cpuinfo | grep processor | tail -1);
	chomp($nconcur);
	$nconcur =~ s/^.*(\d+).*$/$1/;
	$nconcur++;
}
else {
	$nconcur = 1;
	$logger->warn("Unable to detect no# of cpu's for platform \"$osname\", Defaulting to $nconcur parallel job(s)");
}

$nconcur = $opt{'j'} if (exists($opt{'j'}) && $opt{'j'} > 0);

$logger->info("Using $nconcur no# of cpu's for platform \"$osname\"");

my $t0 = [gettimeofday];
my $START_TIME = gmtime;

$logger->info("Batch job starting");

my($PNGDIR) = RIPE::DNSMon::RRD::get_pngdir();

umask(3);

$logger->debug("Locking $PNGDIR/BATCH");
my $locker = LockFile::Simple->make(-warn => 0, -delay => 3, -max => 300, -hold =>3600, -stale => 1) || _die("Cannot create File Locker");
$locker->lock("$PNGDIR/BATCH") || _die("cannot lock $PNGDIR/BATCH");

# more con/2/ml/incoming/Moy.Drug.Ivan.Lapshin.venient access to second constants
#
my $HOUR = Time::Piece->ONE_HOUR;
my $DAY  = Time::Piece->ONE_DAY;
my $WEEK = Time::Piece->ONE_WEEK;


if ($doupdates) {
	####################	UPDATE

	$logger->info("Update process starting");

	$logger->info("getting monitor list");

	foreach my $af (&AF_INET, &AF_INET6) {

		my @p = (sort &RIPE::DNSMon::RRD::raw_mons($af));

		run_in_parallel( sub {
			my (@_p) = (@_);
			updaterrds(@_p,$af);
		}, @p);
	}

	$logger->info("Update process finished, delta=", tv_interval($t0), "");
}
if ($doplots) {

	################    GENPLOTS DAILY

	if ( ($START_TIME->hour == GENPLOTS_DAILY_HOUR) &&		    # runs once a day
		((gmtime->epoch - $START_TIME->epoch) < TIME_TO_LIVE) ) {   # and not exceeding
		foreach my $af (&AF_INET, &AF_INET6) {
			my $ipv = $af == &AF_INET ? "IPv4" : $af == &AF_INET6 ? "IPv6" : "Unknown";

			eval {							    # allocated time
				my $JOB_START_TIME = [gettimeofday];
				$logger->info("Daily $ipv Plot Generation starting");

				# week / 2 weeks
				#
				run_in_parallel(sub {
					my ($s) = (@_);
					my $lt = Time::Piece->strptime(RIPE::DNSMon::RRD::last_val($s,$af), '%s');
					serplot($s, $lt-$DAY*14-($lt->epoch%$DAY), $lt-($lt->epoch%$DAY)-1, $af);
					serplot($s, $lt-$DAY*7-($lt->epoch%$DAY), $lt-($lt->epoch%$DAY)-1, $af);
				}, (RIPE::DNSMon::RRD::all_servers($af)));
				run_in_parallel(sub {
					my ($d) = (@_);
					my $lt = Time::Piece->strptime(RIPE::DNSMon::RRD::last_val_domain($d, $af), '%s');
					domplot($d, $lt-$DAY*14-($lt->epoch%$DAY), $lt-($lt->epoch%$DAY)-1, $af);
					domplot($d, $lt-$DAY*7-($lt->epoch%$DAY), $lt-($lt->epoch%$DAY)-1, $af);
				}, (RIPE::DNSMon::RRD::all_domains($af)));

				 run_in_parallel(sub {
					my ($m) = (@_);
					my $lt = Time::Piece->strptime(RIPE::DNSMon::RRD::last_val_mon($m, $af), '%s');
					monplot($m, $lt-$DAY*14-($lt->epoch%$DAY), $lt-($lt->epoch%$DAY)-1, $af);
					monplot($m, $lt-$DAY*7-($lt->epoch%$DAY), $lt-($lt->epoch%$DAY)-1, $af);
				}, (RIPE::DNSMon::RRD::all_mons($af)));

				$logger->info("Daily $ipv Plot Generation finished, delta=", tv_interval($JOB_START_TIME));
			};
			$logger->fatal("Daily $ipv Plot Generation failed: \"$@\"") if $@;
		}
	}


	######################### GENPLOTS HOURLY

	if ( (gmtime->epoch - $START_TIME->epoch) < TIME_TO_LIVE ) {	# do not run if
		foreach my $af (&AF_INET, &AF_INET6) {
			my $ipv = $af == &AF_INET ? "IPv4" : $af == &AF_INET6 ? "IPv6" : "Unknown";

			eval {							# previous jobs took a long
				my $JOB_START_TIME = [gettimeofday];
				$logger->info("Hourly $ipv Plot Generation starting");

				# 24h / 48h

				 run_in_parallel(sub {
					my ($d) = (@_);
					my $lt = Time::Piece->strptime(RIPE::DNSMon::RRD::last_val_domain($d,$af), '%s');
					$logger->info("domplot($d, $lt-$DAY, $lt-1, $ipv)");
					domplot($d, $lt-$DAY, $lt-1, $af);
				}, (RIPE::DNSMon::RRD::all_domains($af)));

				run_in_parallel(sub {
					my ($s) = (@_);
					my $lt = Time::Piece->strptime(RIPE::DNSMon::RRD::last_val($s,$af), '%s');
					$logger->info("serplot($s, $lt-$DAY, $lt-1, $ipv)");
					serplot($s, $lt-$DAY, $lt-1, $af);
				}, (RIPE::DNSMon::RRD::all_servers($af)));

				run_in_parallel(sub {
					my ($d) = (@_);
					my $lt = Time::Piece->strptime(RIPE::DNSMon::RRD::last_val_domain($d,$af), '%s');
					$logger->info("domplot($d, $lt-2*$DAY, $lt-1, $ipv)");
					domplot($d, $lt-2*$DAY, $lt-1, $af);
				}, (RIPE::DNSMon::RRD::all_domains($af)));

				$logger->info("Hourly $ipv Plot Generation finished, delta=", tv_interval($JOB_START_TIME));
			};
			$logger->fatal("Hourly $ipv Plot Generation failed: \"$@\"") if $@; 
		}
	}
}
$locker->unlock("$PNGDIR/BATCH") || $logger->warn("cannot unlock $PNGDIR/BATCH");
$logger->info("Batch job finished, total delta=", gmtime->epoch - $START_TIME->epoch, "");

END {
    close STDOUT or _die("Error closing STDOUT: $!");
    close STDERR or _die("Error closing STDERR: $!");
}

############################################################

sub report {
    my($kind, $name, $start, $stop, $e, $stime) = @_;

    if ($e =~ /^C /) { # cached
	$logger->info(sprintf("plot_$kind($name, %s, %s): exists",
		$start->datetime, $stop->datetime));
    }
    elsif ($e =~ /^F /) { # just made
	$logger->info(sprintf("plot_$kind($name, %s, %s): generated, %f seconds",
		$start->datetime, $stop->datetime, tv_interval($stime)));
    }
    elsif ($e !~ /^F /) {
	$logger->error("plot_$kind($name) generated ERROR: $e");
    }
}

sub domplot {
    my($domain, $start, $stop, $af) = @_;
    my($e);

    my $stime = [gettimeofday];
    my %a = ( start => $start, stop => $stop, domain => $domain, af => $af);
    $e = RIPE::DNSMon::RRD::plot_domain(%a);
    report("domain", $domain, $start, $stop, $e, $stime);
}

sub serplot {
    my($server, $start, $stop, $af) = @_;
    my($e);

    my $stime = [gettimeofday];
    my %a = ( start => $start, stop => $stop, srv => $server, af => $af);
    $e = RIPE::DNSMon::RRD::plot_server(%a);
    report("server", $server, $start, $stop, $e, $stime);
    $stime = [gettimeofday];
    $e = RIPE::DNSMon::RRD::plot_server_drops(%a);
    report("server_drops", $server, $start, $stop, $e, $stime);
}

sub monplot {
    my($mon, $start, $stop, $af) = @_;
    my($e);

    my $stime = [gettimeofday];
    my %a = ( start => $start, stop => $stop, mon => $mon, af => $af);
    $e = RIPE::DNSMon::RRD::plot_mon(%a);
    report("mon", $mon, $start, $stop, $e, $stime);
    $stime = [gettimeofday];
    $e = RIPE::DNSMon::RRD::plot_mon_drops(%a);
    report("mon_drops", $mon, $start, $stop, $e, $stime);
}

# XXX Use Log4Perl here

sub _die
{
    my($message) = @_;
    $logger->logdie($message, Carp::longmess());
}

sub updaterrds(\@$) {
	my ($p_ref,$af) = @_;
	my $ipv = $af == &AF_INET ? "IPv4" : $af == &AF_INET6 ? "IPv6" : "Unknown";

	foreach my $p (@$p_ref) {

		$logger->info("monitor: $p");

		my $d = Time::Piece->strptime(RIPE::DNSMon::RRD::probe_last_data($p,$af), "%s");
		my $r = Time::Piece->strptime(RIPE::DNSMon::RRD::probe_last_rrd($p,$af), "%s");

		next if ($d->epoch == 0);
		if ($d-$r != 0) {
			$logger->info("$p: data=$d, rrd=$r, diff=", $d - $r, "");

			my @todo = RIPE::DNSMon::RRD::probe_files_todo($p,$af);
			$logger->info("$p: do_updates(@todo,$ipv)");
			RIPE::DNSMon::RRD::do_updates(@todo,$af);
		}
	}
}

sub waitforchild {

	my $error;

	$logger->debug("waiting for process ...");
	my $pid = wait();
	$logger->info("found pid $pid, delta=", exists($child_t{$pid}) ? tv_interval($child_t{$pid}) : "UNKNOWN");
	delete $child_t{$pid} if exists($child_t{$pid});

	if ($pid == -1) {
		$logger->fatal("No child process to wait for?");
		exit 1;
	}
	else {
		if ($? != 0) {	# errors
			if ($? > 128) {  # killed by UNIX signal 
				my $sig = $? & 127;
				$error="received signal: " . `kill -l $sig`;
			}
			else {	# non-zero return code from program
				$error="exit status: $?";
			}

			$logger->fatal("updaterrds() terminated abnormally: $error");
		}
		$nrunning -= 1;
	}
}

sub fork_run_in_parallel ($@) {
	my($func,$p) = @_;
	my $pid;
	if ($nrunning >= $nconcur) {
		waitforchild();
	}
	if ($pid = fork) {
		# parent
		$child_t{$pid} = [gettimeofday];
		$nrunning += 1;
		$logger->debug("forked pid $pid, $nrunning procs running");
		$logger->debug("Processing for $p");
	}
	elsif (defined $pid) {  
		# $pid is zero here if defined; this is the child process.
		&$func($p);
		exit(0);
	}
	else {
		$logger->logdie("cannot fork ($!)");
	}
}

sub run_in_parallel (&@) {
	my($func,@p) = @_;
	while($#p > -1) {
		my $p = shift(@p);
		fork_run_in_parallel($func,$p);
		#thread_run_in_parallel($func,$p)
	}

	while ($nrunning > 0 ) {
		waitforchild();
	}
}

sub HELP_MESSAGE() {
        &usage;
}

sub usage {
        print STDERR <<"EOT";
$0 [-h] | [-p] | [-u] | -j n
	-j n use n count parallel jobs. defaults to CPU cores autodetected
	-p   only plots 
	-u   only update rrd's
	-h   this message
EOT
        exit 1;
}

