package RIPE::DNSMon::DNSMonConfig;
############################################################################
###    (C)opyright 2003 - 2008 RIPE NCC
###    This file is part of DNSMon
###
###    DNSMon is free software: you can redistribute it and/or modify
###    it under the terms of the GNU General Public License as published by
###    the Free Software Foundation, either version 3 of the License, or
###    (at your option) any later version.
###
###    DNSMon is distributed in the hope that it will be useful,
###    but WITHOUT ANY WARRANTY; without even the implied warranty of
###    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
###    GNU General Public License for more details.
###
###    You should have received a copy of the GNU General Public License
###    along with DNSMon.  If not, see <http://www.gnu.org/licenses/>.
############################################################################

use File::stat;
use Time::localtime;
use Log::Log4perl 'get_logger';
use Socket;
use Socket6;
use strict;


our $VERSION = 1.13;

## TEST

my $file_path;

#this should be ideally static. But it happens that apache forks
#new processes so that it gets resetted. The file actually gets read everytime another
#thread is spawned
my $last_tm;
our $log = get_logger();

sub new {

  my $class = shift;
  $file_path = shift;
  $last_tm = -1;
  my $self = {};
  init_maps();
	
  return bless $self, $class;
}


# my $file_path = '/home/fvivoli/Documents/projects/dnsmon/conf/config'; 	
# my $file_path = '/tmp/config';
#$date_string = ctime(stat($file_path)->mtime);
#print "file $file_path updated at $date_string\n"; 




my %seralias;	# alias names for servers
my %monalias;	# alias names for probes
my %sername;	# printable names for servers
my %monname;	# printable names for probes
my %domains;	# IPv4 servers in each domain
my %domains6;	# IPv6 servers in each domain
my %addrs;	# IPv4 addresses of nameservers
my %addrs6;	# IPv6 addresses of nameservers



##Refactor accessors with something cleverer
sub get_domains($){
	my $self = shift;
	my ($af) = @_;
	$self->read_config();
	return $af == &AF_INET ? \%domains : \%domains6;
}

#get servers configured for a domain
sub get_domainkeys($$){
	my $self = shift;
	my ($domain,$af) = @_;
	my $domains = $self->get_domains($af);
	return $$domains{$domain};
}

# Get address for nameserver
sub get_addrs($){
	my $self = shift;
	my ($af) = @_;
	$self->read_config();
	return $af == &AF_INET ? \%addrs : \%addrs6;
}

sub get_addrkeys($$){
	my $self = shift;
	my ($addr,$af) = @_;
	my $addrs = $self->get_addrs($af);
	return $$addrs{$addr};
}

sub get_monnames(){
	my $self = shift;
	$self->read_config();
	return \%monname;
}

sub get_sernames(){
	my $self = shift;
	$self->read_config();
	return \%sername;
}

sub get_monalias(){
	my $self = shift;
	$self->read_config();
	return \%monalias;
}

sub get_seralias(){
	my $self = shift;
	$self->read_config();
	return \%seralias;
}

sub get_seraliaskeys($){
	my $self = shift;
	my $alias = shift;
	my $seralias = $self->get_seralias();
	return $seralias{$alias};
}

sub init_maps(){

	%seralias = ();	# alias names for servers
	%monalias = ();	# alias names for probes
	%sername = ();	# printable names for servers
	%monname = ();	# printable names for probes
	%domains = ();	# IPv4 servers in each domain	
	%domains6 = ();	# IPv6 servers in each domain	
	%addrs = (); # IPv4 nameserver addresses
	%addrs6 = (); # IPv6 nameserver addresses

}


sub read_config {
	my $f = $file_path;	
	my $log = get_logger();
	
	if(file_updated()){
		#$log->info("Re-reading config file...\n"); 
		# read config file, store in globals
		init_maps();
		open(CONFIG, $f) || die("open config file: $f");
		
		
		while(<CONFIG>) {
			if (/^seralias\s+(\S+)\s+(.*)$/) {
			$seralias{$1} = $2;
			}
			if (/^sername\s+(\S+)\s+(.*)$/) {
			$sername{$1} = $2;
			}
			if (/^tbalias\s+(\S+)\s+(.*)$/) {
			$monalias{$1} = $2;
			}
			if (/^tbname\s+(\S+)\s+(.*)$/) {
			$monname{$1} = $2;
			}
			if (/^domain\s+(\S+)\s+(.*)$/) {
			$domains{$1} = $2;
			$domains{$1} =~ tr/[A-Z]/[a-z]/;
			}
			if (/^domain6\s+(\S+)\s+(.*)$/) {
			$domains6{$1} = $2;
			$domains6{$1} =~ tr/[A-Z]/[a-z]/;
			}
			if (/^addr\s+(\S+)\s+(.*)$/) {
			$addrs{$1} = $2;
			$addrs{$1} =~ tr/[A-Z]/[a-z]/;
			}
			if (/^addr6\s+(\S+)\s+(.*)$/) {
			$addrs6{$1} = $2;
			$addrs6{$1} =~ tr/[A-Z]/[a-z]/;
			}
		}
		close(CONFIG);
	}
	
}


sub file_updated(){
	my $self = shift;
	my $st;
	my $ret = 0;

	if ($st = stat($file_path)) {
		my $tm = $st->mtime;
		$log->debug("Last TM: $last_tm Current TM: $tm");	
		if( $last_tm==-1 || $tm>$last_tm){
			$last_tm = $tm;
			$ret = 1;
		}
	} else {
		$log->error("unable to check modification time of $file_path! $!");	
	}
	return $ret;
}
1;
