package RIPE::DNSMon::RRD;
############################################################################
###    (C)opyright 2003 - 2008 RIPE NCC
###    This file is part of DNSMon
###
###    DNSMon is free software: you can redistribute it and/or modify
###    it under the terms of the GNU General Public License as published by
###    the Free Software Foundation, either version 3 of the License, or
###    (at your option) any later version.
###
###    DNSMon is distributed in the hope that it will be useful,
###    but WITHOUT ANY WARRANTY; without even the implied warranty of
###    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
###    GNU General Public License for more details.
###
###    You should have received a copy of the GNU General Public License
###    along with DNSMon.  If not, see <http://www.gnu.org/licenses/>.
############################################################################


=head1 NAME

RIPE::DNSMon::RRD - 

=head1 SYNOPSIS 

use RIPE::DNSMon::RRD

=head1 DESCRIPTION

dnsmon perl code

Most of the perl code is concentrated in this package.
The idea is not to have any real programs in the mason 
pages. 

This is definitely still prototype code; e.g. the result
of experimenting with both the external presentation
of results *and* the ways of coding it.

It can definitely be improved, especially in:

=over

=item naming and structure of the graphics store.

=item normalising the representation and handling of time, I changed to L<Time::Piece> half way through and did not retrofit everywhere.

=item more general use of configuration variables.

=back

=head3 Prerequisites for Understanding

undestanding of dnsmon web site and documentation

=over

=item B<mason>

=item B<rrd>

=back

=head3 Terminology 

=over

=item B<server> 

a DNS name server being monitored

=item B<domain>

A set of servers typically those authoritative for a particular DNS domain

=item B<probe>

An entity sending test traffic to a number of servers currently these are TTM test-boxes currently all probes monitor all servers

=item B<monitor>

Same as probe (historical)

=back

=cut


use strict;
use lib '/usr/local/rrdtool/lib/perl';
use AppConfig;
use LockFile::Simple;
use Time::Piece;
use RRDs;
use RIPE::DNSMon::DNSMonConfig;
use Log::Log4perl 'get_logger';
use Time::HiRes qw(gettimeofday tv_interval);
use Cache::FileCache;
use File::Basename;
use File::Find;
use File::Path;

# Access to AF_INET/AF_INET6 and various IPv4/IPv6 address manipulation functions
use Socket;
use Socket6;

use constant SEEK_SET => 0;

BEGIN {
    use Exporter   ();
    our ($VERSION, @ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);

    # set the version for version checking
    $VERSION = do { my @r = (q$Revision: 1.14.2.1 $ =~ /\d+/g); sprintf "%d."."%02d" x $#r, @r }; # must be all one line, for MakeMaker

    @ISA         = qw(Exporter);
    @EXPORT      = ( );
    %EXPORT_TAGS = ( );     # eg: TAG => [ qw!name1 name2! ],

    # your exported package globals go here,
    # as well as any optionally exported functions
    @EXPORT_OK   = ( );
}
our @EXPORT_OK;

                                              
                                              
our $log = get_logger();


############  C o n f i g u r a t i o n  ############

=head1 CONFIGURATION

=cut

our $CONFIGFILENAME = $ENV{DNSMONCONFIG} ||"/usr/local/ncc/etc/dnsmon-rrd.conf";

if (not -e $CONFIGFILENAME) {
    $log->logdie("configuration file $CONFIGFILENAME not found");
}

my $CONFIG = new AppConfig;
$CONFIG->define(
    'RRDDIR=s',
    'RRDDIR6=s',
    'PNGDIR=s',
    'DATADIR=s',
    'CACHEDIR=s',
    'PNGPATH=s',
    'CONFIG=s',
    'NOGRAPH=s',
);
$CONFIG->file($CONFIGFILENAME) or 
    $log->logdie("ERROR: reading configuration file $CONFIGFILENAME");

my $RRDDIR6 = $CONFIG->RRDDIR6;
my $RRDDIR = $CONFIG->RRDDIR;
my $PNGDIR = $CONFIG->PNGDIR;
my $DATADIR = $CONFIG->DATADIR;
my $CACHEDIR = $CONFIG->CACHEDIR;
my $PNGPATH = $CONFIG->PNGPATH;
my $NOGRAPH = $CONFIG->NOGRAPH;

my $cache = new Cache::FileCache( { 'namespace' => 'DNSMon_RRD', 'default_expires_in' => 1800, auto_purge_on_set => 0, cache_root => $CACHEDIR });
my $offset_cache = new Cache::FileCache( { 'namespace' => 'DNSMon_RAW', 'default_expires_in' => 7* 24 * 3600, auto_purge_on_set => 1, cache_root => $CACHEDIR });

# To enable a flat png directory structure set to 1 (old behaviour)
my $opt_flatpngdir=0;

# XXX Transition hack, remove after 2 weeks when activated
# It will allow for the png's to be still found in the legacy location as if
# $opt_flatpngdir=1. 
# This will improve responsiveness while transitioning to the new layout
my $opt_flatpngdir_transition_kludge=1; 

# XXX Don't enable, but leave the code in place when a suitable and fast caching mechanism appears
my $enable_cache = 0;

#create wrapper object for hashes
#$CONFIG->CONFIG is the path of the web configuration file
my $dns_config = new RIPE::DNSMon::DNSMonConfig($CONFIG->CONFIG);

# verbosity
#
our $opt_d = 0;
our $opt_v = 0;


### PLOT Defaults

my $MINVALS = 15;        # minimum number of valid probes required to make plot
my $MINPROBEVAL = 4;

my $GREEN    = 'A0FFA0';    # colours
my $YELLOW    = 'FFFF00';
my $ORANGE    = 'FFCC00';
my $RED        = 'FF0000';
my $LGTBLUE    = '80C0FF';
my $BLUE    = '0000FF';
my $WHITE    = 'FFFFFF';
my $BLACK    = '000000';

my %DEFS_DOM_PLOT = (
    cf => 'AVERAGE',    # consolidation function
    maxpct => 50,        # y axis max value 
    valid => 1,        # maximum percentage of invalid data to include a probe
    minmons => 5,        # minimum number of probes for a plot
    orange  => '66:90',    # orange color range
    red => '90',        # red color minimum
    opt_w => 800,        # plot width in pixels
    opt_h => 500,        # plot height in pixels
    opt_v => 0,        # verbose switch
    fuzzy => 0,        # fuzzy matching
	af => &AF_INET,		# Address family we want to plot for
);

my %DEFS_SRV_PLOT = (
    cf => 'AVERAGE',    # consolidation function
    maxms => 250,        # y axis max value
    valid => 10,        # maximum percentage of invalid data to include a probe
    minmons => 5,        # minimum number of probes for a plot
    orange  => '66:90',    # orange color range
    red => '90',        # red color minimum
    opt_w => 800,        # plot width in pixels
    opt_h => 500,        # plot height in pixels
    opt_v => 0,        # verbose switch
    fuzzy => 0,        # fuzzy matching
	af => &AF_INET,		# Address family we want to plot for
);

my %DEFS_SRV_PLOT_DROPS = (
    cf => 'AVERAGE',    # consolidation function
    maxpct => 33,        # y axis max value
    valid => 10,        # maximum percentage of invalid data to include a probe
    minmons => 5,        # minimum number of probes for a plot
    orange  => '66:90',    # orange color range
    red => '90',        # red color minimum
    opt_w => 800,        # plot width in pixels
    opt_h => 500,        # plot height in pixels
    opt_v => 0,        # verbose switch
    fuzzy => 0,        # fuzzy matching
	af => &AF_INET,		# Address family we want to plot for
);

my %DEFS_MON_PLOT = (
    cf => 'AVERAGE',    # consolidation function
    maxms => 250,        # y axis max value
    valid => 10,        # maximum percentage of invalid data to include a probe
    orange  => '66:90',    # orange color range
    red => '90',        # red color minimum
    opt_w => 800,        # plot width in pixels
    opt_h => 500,        # plot height in pixels
    opt_v => 0,        # verbose switch
    fuzzy => 0,        # fuzzy matching
	af => &AF_INET,		# Address family we want to plot for
);

my %DEFS_MON_PLOT_DROPS = (
    cf => 'AVERAGE',    # consolidation function
    maxpct => 33,        # y axis max value
    valid => 10,        # maximum percentage of invalid data to include a probe
    orange  => '66:90',    # orange color range
    red => '90',        # red color minimum
    opt_w => 800,        # plot width in pixels
    opt_h => 500,        # plot height in pixels
    opt_v => 0,        # verbose switch
    fuzzy => 0,        # fuzzy matching
	af => &AF_INET,		# Address family we want to plot for
);


### RRD specific stuff

# XXX when these change, update documentation for rrdcreate!

my @rras = (    # structure of the RRDBs
    '0.5:1:43200', # 30 days of 1/min
    '0.5:5:28800', # 100 days of 1/5min
    '0.5:60:9600', # 400 days of 1/h
); 

my @cfs = (    # consolidation functions for each variable
    'AVERAGE',
    'MIN',
    'MAX',
);

my $STEP =     60;        # rrd step in seconds
my $HEARTBEAT =    180;        # rrd heartbeat in seconds
my $STARTT =    1059609600;    # archive start time (31.7.03)

my $RRDTOOL_AT_TIME_FMT = '%H:%M %d.%m.%Y';

our $LASTTROUNDDOWN = 3600;

#####################################################

my $locker = LockFile::Simple->make(-warn => 0, -delay => 3, -max => 100, -hold =>600, -stale => 1) || $log->logdie("RIPE::DNSMon::RRD: Cannot create File Locker");

# from the config file
# my %seralias = ();    # alias names for servers
# my %monalias = ();    # alias names for probes
# my %sername = ();    # printable names for servers
# my %monname = ();    # printable names for probes
# my %domains = ();    # servers in each domain

# read_config();

=head1 FUNCTIONS

=over

=item get_pngdir

export PNGDIR    

=cut

sub get_pngdir    {
    return($PNGDIR);
}

=item get_pngpath

export component path of PNGDIR

=cut

sub get_pngpath    {
    return($PNGPATH);
}

=item all_mons

=head3 NAME

all_mons -- return list of all known probes

=head3 SYNOPSIS

C<all_mons(I<$af>)>

=head3 DESCRIPTION

Return a list of all known probes for the given address family

=over

=item I<$af>

Address family this probe covers, either AF_INET (default) or AF_INET6

=back

=head3 RETURN VALUES

=over

=item I<@probe_list>

List of probes available for this protocol.

=back

=cut

sub all_mons {
	my ($af) = @_;
	my $rrddir = $af == &AF_INET6 ? $RRDDIR6 : $RRDDIR;
    my(@mons) = ();
    opendir(DIR, "$rrddir") || $log->logdie("opendir($rrddir) failed");
    @mons = grep(!/^\./, readdir(DIR));
    close(DIR);
    return(sort @mons);
}

=item raw_mons

=head3 NAME

raw_mons -- return list of all probes with raw data for the given address family

=head3 SYNOPSIS

C<raw_mons(I<$af>)>

=head3 DESCRIPTION

return list of all probes with raw data for the given address family.

=over 

=item I<$af>

Currently I<$af> can be either AF_INET or AF_INET6. These constants are loaded
from Socket and Socket6. When unspecified defaults to AF_INET.

=back

=head3 RETURN VALUES

=over

=item I<@probe_list>

list of usable probes

=back

=cut

sub raw_mons {
	my($af) = @_;
    my(@mons) = ();
    my(%seen) = ();

	my $fprefix = $af == &AF_INET6 ? 'DNS6' : 'DNS';

    opendir(DIR, "$DATADIR") || $log->logdie("opendir($DATADIR) failed");
    while ($_ = readdir(DIR)) {
        if (/${fprefix}\.(.*)\.[0-9]+$/) {
            next if $seen{$1};
            $seen{$1} = 1;
            push(@mons, $1);
        }
    }
    close(DIR);
    return(@mons);
}

=item all_domains

=head3 NAME

all_domains -- return list of all known domains

=head3 SYNOPSIS

C<all_domains(I<$af>)>

=head3 DESCRIPTION

Return the list of all domains for the given address family

=over

=item I<$af>

Address family to filter on. Either AF_INET (the default) or AF_INET6

=back

=head3 RETURN VALUES

=over

=item I<@domainlist>

list of domains which are monitored in the given address family

=back

=cut

sub all_domains {
	my ($af) = @_;
    my(@mons) = ();
    my $domains = $dns_config->get_domains($af); 
    return(sort keys %$domains);
}

=item all_servers

=head3 NAME

all_servers -- return list of all known servers

=head3 SYNOPSIS

C<all_servers(I<$af>)>

=head3 DESCRIPTION

Return a list of all know nameservers for the given address family

=over

=item I<$af>

Address family to filter on, either AF_INET (the default) or AF_INET6

=head3 RETURN VALUES

=over

=item I<@serverlist>

list of servers that are being monitored for the given address family

=back

=cut

sub all_servers {
	my ($af) = @_;
	my $rrddir = $af == &AF_INET6 ? $RRDDIR6 : $RRDDIR;
    my (@servers) = ();

	if (opendir(DIR,"$rrddir")) {
		foreach my $probedir (grep { -d "$rrddir/$_" && !/^\./ } readdir(DIR)) {
			if (opendir(PROBEDIR,"$rrddir/$probedir")) {
				push (@servers, grep { -f "$rrddir/$probedir/$_" && !/^\./}  readdir(PROBEDIR));
				closedir(PROBEDIR);
			}	
		}
		closedir(DIR);
	}
	@servers = keys %{{map {$_ => 1} @servers}};

    return(sort domain_rcomp  @servers);
}

=item mon_servers

=head3 NAME

mon_servers -- return list of all servers with data from this probe

=head3 SYNOPSIS

C<mon_servers(I<$probe>,I<$af>)>

=head3 DESCRIPTION

Return a list of all known servers for the given probe and address family

=over 

=item I<$probe>

probe we want to see the server coverage from

=item I<$af>

Address family this probe covers, either AF_INET (default) or AF_INET6

=back

=head3 RETURN VALUES

=over

=item I<@server_list>

List of servers available for this probe and protocol.

=back

=cut

sub mon_servers {
    my($m,$af) = @_;
	my $rrddir = $af == &AF_INET6 ? $RRDDIR6 : $RRDDIR;
    my (@servers) = ();

    return(@servers) unless (-d "$rrddir/$m");

    opendir(DIR, "$rrddir/$m") || $log->logdie("opendir($rrddir/$m)");
    while ($_ = readdir(DIR)) {
        next if /^\./;
        push @servers, $_;
    }
    closedir(DIR);
    return(sort @servers);
}


=item known_mons

=head3 NAME

known_mons -- return list of all valid probes in argument list

=head3 SYNOPSIS

C<known_mons(I<@probes>,I<$af>)>

=head3 DESCRIPTION

Filters the argument list for valid probes for the given address family

=over

=item I<@probes>

List of probes to be filtered. If a probe has an alias it will be resolved here.

=item I<$af>

Address family to filter on. Either AF_INET (default) or AF_INET6

=back

=head3 RETURN VALUES

=over

=item I<@probes>

List of valid probes with any alias resolved

=back

=cut

sub known_mons(\@$) {
	my ($probe_ref, $af) = @_;
    my(@mons) = ();

	$log->logdie("not a reference") unless ref($probe_ref);

    foreach my $a (@$probe_ref) {
        $a = &mon_alias($a);
        if (&known_mon($a,$af)) {
            push(@mons, $a);
        }
        else {
            $log->info("$a is not a valid probe, ignored");
        }
    }
    return(@mons);
}


=item argv_servers

return list of all known servers in arg list

=cut

sub argv_servers {
    my ($af) = @_;
    my(@servers) = ();
    my($a) = '';
    foreach $a (@_) {
        $a = &ser_alias($a);
        if (&known_server($a,$af)) {
            push(@servers, $a);
        }
        else {
            $log->info("$a is not a valid server, ignored");
        }
    }
    return(@servers);
}


=item known_mon

=head3 NAME

known_mon -- return true if arg is known probe

=head3 SYNOPSIS

C<known_mon(I<$probe>,I<$af>)>

=head3 DESCRIPTION

Check wether the given probe is valid for the given address family

=over

=item I<$probe>

Probe to check for

=item I<$af>

Address family for probe. Either AF_INET (default) or AF_INET6

=back

=head3 RETURN VALUES

=over

=item I<$valid>

true if probe is valid

=back

=cut

sub known_mon {
    my($m,$af) = @_;
	my $rrddir = $af == &AF_INET6 ? $RRDDIR6 : $RRDDIR;
    return(1) if (-d "$rrddir/$m");
    return(0);
}

=item known_server

=head3 NAME

known_server -- return number of probes with data for this server

=head3 SYNOPSIS

C<known_server(I<$server>,I<$af>)>

=head3 DESCRIPTION

return number of probes with data for this server and address family

=over

=item I<$server>

nameserver we want to see the amount of probes from

=item I<$af>

address family for probes, either AF_INET (default) or AF_INET6

=back

=head3 RETURN VALUES

=over

=item I<$probecount>

Amount of probes having data for this server

=back

=cut

sub known_server {
    my($s,$af) = @_;
    my(@mons) = ();
    my($nmons) = 0;
    my($m) = 0;
	my $rrddir = $af == &AF_INET6 ? $RRDDIR6 : $RRDDIR;

    opendir(DIR, $rrddir) || $log->logdie("opendir: $rrddir");
    @mons = grep(!/^\./, readdir(DIR));
    close(DIR);
    foreach $m (@mons) {
        $nmons++ if (-f "$rrddir/$m/$s");
    }
    return($nmons);
}


=item domain_servers

=head3 NAME

domain_servers -- return list of known servers for a domain

=head3 SYNOPSIS

C<domain_servers(I<$domain>,I<$af>)>

=head3 DESCRIPTION

Retrieve the list of nameservers associated with this domain for the given address family.

=over 

=item I<$domain>

domain we want to see the server coverage from

=item I<$af>

Address family for the covering servers, either AF_INET (default) or AF_INET6

=back

=head3 RETURN VALUES

=over

=item I<@server_list>

List of servers available for this domain and protocol.

=back

=cut

sub domain_servers {
    my($domain,$af) = @_;
    my(@known_servers) = ();

    $log->warn("domain_servers('$domain'): unknown domain") unless $dns_config->get_domainkeys($domain,$af);

    my @servers = split(/[ \t\n]+/, $dns_config->get_domainkeys($domain, $af));
    foreach my $s (sort domain_rcomp  @servers) {
        $s = ser_alias($s);
        push(@known_servers, $s) if known_server($s,$af);
    }
    return(@known_servers);
}

=item server_domains

=head3 NAME
=head3 SYNOPSIS
=head3 DESCRIPTION
=head3 RETURN VALUES

=cut

sub server_domains {
	my ($server,$af) = @_;
	my (@domains);
	foreach my $domain (all_domains($af)) {
		push(@domains, $domain) if (grep {$_ eq $server} RIPE::DNSMon::RRD::domain_servers($domain,$af));
	}
	return @domains;
}


=item mknames(@things)

Makes shortest possible set of uppercase alpha names for @things.

Returns: %names - assoc name translation array

=cut

sub mknames {
    my(@things) = @_;
    my(%names) = (); 
    my($t);
    my($n) = 'A';

    foreach $t (@things) {
        $names{$t} = $n++;
    }
    return(%names);
}


=item select_mons

=head3 NAME

select_mons -- Selects usable @probes for a server

=head3 SYNOPSIS

C<select_mons(I<$start>, I<$stop>, I<$threshold>, I<$server>, I<$af>, I<@probes>)>

=head3 DESCRIPTION

Selects usable @probes for I<$server> between I<$start> and I<$stop> times based
on minimum I<$threshold> percent available data. 

=over

=item I<$start>

Start time, in rrdtool '%H:%M %d.%m.%Y' format

=item I<$stop>

End time, in rrdtool '%H:%M %d.%m.%Y' format

=item I<$threshold>

threshold, in percent, that this probe should have available minimum amount of data for the given server.

=item I<$server>

server for we want to get the list of valid (enough data) probes for.

=item I<$af>

Address family this server covers, either AF_INET (default) or AF_INET6

=item I<@probes>

Probes available for filtering.

=back

=head3 IMPLEMENTATION NOTES

Threshold only applied above the baseline minimum threshold in order to exclude failures of all probes.

=head3 RETURN VALUES

=over

=item I<@usable_mons>

list of usable probes

=back

=head3 BUGS

This is a major drag on performance! Caching this info could help.

Should merge code with select_servers

=cut
sub select_mons {
    my($start, $stop, $threshold, $server, $af, @mons) = @_;
    my(@selected) = ();    # selected probes
    my(%ratio) = ();    # valid / invalid
    my($minratio) = 1;    # minimum ratio found 
    my($m);
    my(%hasdata) = ();    # server has data
	my $rrddir = $af == &AF_INET6 ? $RRDDIR6 : $RRDDIR;

    foreach $m (@mons) {                # for each probe
	my($rrd_start_t,$step,$names,$array, $rrd_lastmod, $start_s, $stop_s, $cache_object, @rrdstat);

	$start_s = Time::Piece->strptime($start, $RRDTOOL_AT_TIME_FMT);
	$stop_s = Time::Piece->strptime($stop, $RRDTOOL_AT_TIME_FMT);

        next unless -d "$rrddir/$m";
        next unless -r "$rrddir/$m/$server";
	@rrdstat = stat(_);

	($rrd_start_t,$step,$names,$array, $rrd_lastmod) = @{ $cache_object } if ($enable_cache && ($cache_object = $cache->get("$m-$server-AVERAGE-@{[$start_s->epoch]}-@{[$stop_s->epoch]}")));

	$log->debug("Using cache entry \"$m-$server-AVERAGE-@{[$start_s->epoch]}-@{[$stop_s->epoch]}\"") if defined($rrd_lastmod);

	if (not defined($rrd_lastmod) || (defined($rrd_lastmod) && $rrd_lastmod < $rrdstat[9] )) {
		my $t1 = [gettimeofday];

		$rrd_lastmod = $rrdstat[9];

		$log->debug("RRDs::fetch \"$rrddir/$m/$server\", \"AVERAGE\", \"-s $start\", \"-e $stop\"");
		($rrd_start_t,$step,$names,$array) = RRDs::fetch "$rrddir/$m/$server", "AVERAGE", "-s $start", "-e $stop";
		$log->debug("RRDs::fetch returned ($rrd_start_t,$step,$names,$array) delta=",tv_interval($t1));
		my $error = RRDs::error;
		if ($opt_d) {
			$log->error("RRDs::error returned: \"$error\"") if $error;
		}
		next if (defined($error) && $error ne '');
		if ($enable_cache) {
			$cache->set("$m-$server-AVERAGE-@{[$start_s->epoch]}-@{[$stop_s->epoch]}", [$rrd_start_t,$step,$names,$array, $rrd_lastmod] );
			$log->debug("stored cache entry: $m-$server-AVERAGE-@{[$start_s->epoch]}-@{[$stop_s->epoch]}");
		}
	}

        my($undefs) = 0;
        my($lines) = 0;
        foreach my $line (@$array) {    # go thru all data for period
            if (!defined($line) || ((!defined(${$line}[0]) || ${$line}[0] eq '' ) && (!defined(${$line}[1]) || ${$line}[1] eq ''))) {
                $undefs++;              # number of undefined data points
            }
            $lines++;                   # number of all possible data points
        }
        $log->debug("undefs($m,$server,$af) = $undefs in $lines") if $opt_d;
        $ratio{$m} = $lines > 0 ? $undefs / $lines : 0;  # undefined ratio
        $minratio = $ratio{$m} if $ratio{$m} < $minratio;# minimum ratio of 
                                        # all probes
        $hasdata{$m} = 1;
    }

    foreach $m (@mons) {
        next unless $hasdata{$m};
        if (($ratio{$m}-$minratio) > $threshold/100) { # subtlety: minratio
            next;
        }
        push(@selected, $m);
    }
    return(@selected);
}


=item select_servers

=head3 NAME

select_servers -- Selects usable @servers

=head3 SYNOPSIS

C<select_servers(I<$start>, I<$stop>, I<$threshold>, I<$probe>, I<$af>, I<@servers>)>

=head3 DESCRIPTION

Selects usable @servers for I<$probe> between I<$start> and I<$stop> times based
on minimum I<$threshold> percent available data. 

=over

=item I<$start>

Start time, in rrdtool '%H:%M %d.%m.%Y' format

=item I<$stop>

End time, in rrdtool '%H:%M %d.%m.%Y' format

=item I<$threshold>

threshold, in percent, that this probe should have available minimum amount of data for the given server.

=item I<$probe>

probe for we want to get the list of valid (enough data) servers for.

=item I<$af>

Address family this probe covers, either AF_INET (default) or AF_INET6

=item I<@servers>

Domain servers available for filtering.

=back

=head3 IMPLEMENTATION NOTES

Threshold only applied above the baseline minimum threshold in order to exclude failures of all probes.

=head3 RETURN VALUES

=over

=item I<@usable_servers>

list of usable servers

=back

=head3 SEE ALSO

L</select_mons>

=head3 BUGS

This is a major drag on performance! Caching this info could help.

Should merge code with select_mons

=cut

sub select_servers {
    my($start, $stop, $threshold, $m, $af, @servers) = @_;
    my(@selected) = ();
    my(%ratio) = ();
    my($minratio) = 1;
    my($s);
	my $rrddir = $af == &AF_INET6 ? $RRDDIR6 : $RRDDIR;


    foreach $s (@servers) {
	my($rrd_start_t,$step,$names,$array, $rrd_lastmod, $start_s, $stop_s, $cache_object,@rrdstat);

	$start_s = Time::Piece->strptime($start, $RRDTOOL_AT_TIME_FMT);
	$stop_s = Time::Piece->strptime($stop, $RRDTOOL_AT_TIME_FMT);

        next unless -d "$rrddir/$m";
        next unless -f "$rrddir/$m/$s";
	@rrdstat = stat(_);

	($rrd_start_t,$step,$names,$array, $rrd_lastmod) = @{ $cache_object } if ($enable_cache && ($cache_object = $cache->get("$m-$s-AVERAGE-@{[$start_s->epoch]}-@{[$stop_s->epoch]}")));

	$log->debug("Using cache entry \"$m-$s-AVERAGE-@{[$start_s->epoch]}-@{[$stop_s->epoch]}\"") if defined($rrd_lastmod);

	if (not defined($rrd_lastmod) || (defined($rrd_lastmod) && $rrd_lastmod < $rrdstat[9] )) {
		my $t1 = [gettimeofday];

		$rrd_lastmod = $rrdstat[9];

		$log->debug("RRDs::fetch \"$rrddir/$m/$s\", \"AVERAGE\", \"-s $start\", \"-e $stop\"");
		($rrd_start_t,$step,$names,$array) = RRDs::fetch "$rrddir/$m/$s", "AVERAGE", "-s $start", "-e $stop";
		$log->debug("RRDs::fetch returned ($rrd_start_t,$step,$names,$array) delta=",tv_interval($t1));
		my $error = RRDs::error;
		if ($opt_d) {
			$log->error("RRDs::error returned: \"$error\"") if $error;
		}
		next if (defined($error) && $error ne '' );
		if ($enable_cache) {
			$cache->set("$m-$s-AVERAGE-@{[$start_s->epoch]}-@{[$stop_s->epoch]}", [$rrd_start_t,$step,$names,$array, $rrd_lastmod] );
			$log->debug("stored cache entry: $m-$s-AVERAGE-@{[$start_s->epoch]}-@{[$stop_s->epoch]}");
		}
	}
        my($undefs) = 0;
        my($lines) = 0;
        foreach my $line (@$array) {
            if (!defined($line) || ((!defined(${$line}[0]) || ${$line}[0] eq '' ) && (!defined(${$line}[1]) || ${$line}[1] eq ''))) {
                $undefs++;
            }
            $lines++;
        }
        $log->debug("undefs($m,$s,$af) = $undefs in $lines") if $opt_d;
        $ratio{$s} =  $lines > 0 ? $undefs / $lines : 0;
        $minratio = $ratio{$s} if $ratio{$s} < $minratio;
    }

    foreach $s (@servers) {
        if (($ratio{$s}-$minratio) > $threshold/100) {
            next;
        }
        push(@selected, $s);
    }
    return(@selected);
}

=item select_mons_domain

=head3 NAME

select_mons_domain -- Selects usable @probes for a domain

=head3 SYNOPSIS

C<select_mons_domain(I<$start>, I<$stop>, I<$threshold>, I<$domain>, I<$af>, I<@probes>)>

=head3 DESCRIPTION

Selects usable @probes for I<$domain> between I<$start> and I<$stop> times based
on minimum I<$threshold> percent available data. 

=over

=item I<$start>

Start time, in rrdtool '%H:%M %d.%m.%Y' format

=item I<$stop>

End time, in rrdtool '%H:%M %d.%m.%Y' format

=item I<$threshold>

threshold, in percent, that selected probes should have available minimum amount of data for the given domain.

=item I<$domain>

domain for we want to get the list of valid (enough data) probes for.

=item I<$af>

Address family this server covers, either AF_INET (default) or AF_INET6

=item I<@probes>

Probes available for filtering.

=back

=head3 IMPLEMENTATION NOTES

Threshold only applied above the baseline minimum threshold in order to exclude failures of all probes.

=head3 RETURN VALUES

=over

=item I<@usable_mons_domain>

list of usable probes

=back

=head3 SEE ALSO

L</select_mons>

=head3 BUGS

This is a major drag on performance! Caching this info could help.

=cut

sub select_mons_domain {
    my($start, $stop, $threshold, $domain, $af, @mons) = @_;
    my($nser) = 0;
    my(%moncnt) = ();
    my(@result) = ();

    my @servers = domain_servers($domain, $af);
    foreach my $s (@servers) {
        my @sel = select_mons($start, $stop, $threshold, $s, $af, @mons);
        foreach my $m (@sel) {
            $moncnt{$m}++;
        }
        $nser++;
    }
    foreach my $m (keys %moncnt) {
        push(@result, $m) if $moncnt{$m} == $nser;
    }
    return(sort @result);
}
            

=item num_rcomp

numeric comparator for sorting etc 

=cut

sub num_rcomp { $b <=> $a; }

=item domain_rcomp

domain comparator for sorting etc 

=cut

# XXX make configurable
our @multilevel_tld = qw(uk hk vn);

sub domain_rcomp { 
	my $cmp;

	my ($a_tld,@a) = reverse split(/\./,$a);
	my ($b_tld,@b) = reverse split(/\./,$b);

	@a = grep { defined $_ } splice(@a,(grep {$a_tld eq $_ } @multilevel_tld) ? 1: 0);
	@b = grep { defined $_ } splice(@b,(grep {$b_tld eq $_ } @multilevel_tld) ? 1: 0);

	my ($org_a) = $#a >= 0 ? shift(@a) : $a_tld;
	my ($org_b) = $#b >= 0 ? shift(@b) : $b_tld;

	$cmp = $org_a cmp $org_b ||  $#a <=> $#b;

	if (!$cmp && $#a == $#b && $#a >= 0) {
		do {
			$cmp = shift(@a) cmp shift(@b);
		} while ($#a >= 0 && $#b >= 0 && !$cmp);
	}
	$cmp;
}


=item last_val

=head3 NAME

last_val -- Get time of last data we have for a server with minimum threshold requirements

=head3 SYNOPSIS

C<last_val(I<$server>,I<$af>)>

=head3 DESCRIPTION

Get time of last data we have for a server with at least I<$MINVALS>
probes rounded down to the hour

=over

=item I<$server>

Name of the DNS server that is being monitored

=item I<$af>

Address family this RRD covers, either AF_INET (default) or AF_INET6

=back

The global default $<MINVALS> hold the minimum amount of probes that have valid
data that is required to make a sensible plot

=head3 RETURN VALUES

Time of last data seen, in seconds since the epoch and rounded down to the hour.

=head3 NOTES

Rounding down to the hour might be too restrictive when frequent updates are desired.
This is controlled by I<$LASTTROUNDDOWN> which could be used by applications as a policy tool.

=cut

sub last_val {
    my($server,$af) = @_;
	my $rrddir = $af == &AF_INET6 ? $RRDDIR6 : $RRDDIR;
	my $ipv = $af == &AF_INET ? "IPv4" : $af == &AF_INET6 ? "IPv6" : "Unknown";
    my(@t) = ();
    my ($m, $minidx);
    foreach $m (all_mons($af)) {
        next unless -f "$rrddir/$m/$server";
		# XXX Cache here ?
        push(@t, RRDs::last("$rrddir/$m/$server"));
    }
    if (($#t + 1) < $MINVALS) {
        $log->error("last_val found less than $MINVALS $ipv probes found for $ipv server $server, got @{[ $#t + 1 ]}");
    }
    @t = sort num_rcomp @t;

	# XXX If array is shorter than $MINVALS just return value of highest index available
	# Previously it just used the MINVALS position which means that $MINVALS
	# amount of probes could have newer data, which acts as a kind of percentile level
	
	$minidx = $#t < $MINVALS ? $#t : $MINVALS;
	
    return($t[$minidx] - $t[$minidx] % $LASTTROUNDDOWN);
}


=item last_val_mon

last_val_mon --  Get time of last data we have from the probe above a minimum threshold

=head3 SYNOPSIS

C<last_val_mon(I<$probe>,I<$af>)>

=head3 DESCRIPTION

Get time of last data we have from at least I<$MINVALS>
DNS nameserver for the probe rounded down to next hour

=over

=item I<$probe>

Name of the probe probing nameservers.

=item I<$af>

Address family this RRD covers, either AF_INET (default) or AF_INET6

=back

The global default $<MINVALS> hold the minimum amount of servers that have valid
data that is required to make a sensible plot

=head3 RETURN VALUES

Time of last data seen, in seconds since the epoch and rounded down to the hour.

=head3 NOTES

Rounding down to the hour might be too restrictive when frequent updates are desired.
This is controlled by I<$LASTTROUNDDOWN> which could be used by applications as a policy tool.

=cut

sub last_val_mon {
    my($mon,$af) = @_;
    my(@t) = ();
    my ($s,$minidx);
	my $rrddir = $af == &AF_INET6 ? $RRDDIR6 : $RRDDIR;
	my $ipv = $af == &AF_INET ? "IPv4" : $af == &AF_INET6 ? "IPv6" : "Unknown";


    foreach $s (mon_servers($mon, $af)) {
        next unless -f "$rrddir/$mon/$s";
        push(@t, RRDs::last("$rrddir/$mon/$s"));
    }
    if ($#t + 1 < $MINVALS) {
        $log->error("last_val_mon found less than $MINVALS $ipv servers found for $ipv probe $mon, got @{[ $#t + 1 ]}");
    }
    @t = sort num_rcomp @t;

	# XXX If array is shorter than $MINVALS just return value of highest index available
	# Previously it just used the MINVALS position which means that $MINVALS
	# amount of probes could have newer data, which acts as a kind of percentile level
	
	$minidx = $#t < $MINVALS ? $#t : $MINVALS;

    return($t[$minidx] - $t[$minidx] % $LASTTROUNDDOWN);
}


=item last_val_domain

get time of last data we have for domain 
($MINVALS probes/server, down to next hour)

=head3 NAME

last_val_domain -- Get time of last data we have for a domain with minimum threshold requirements

=head3 SYNOPSIS

C<last_val_domain(I<$domain>,I<$af>)>

=head3 DESCRIPTION

Get time of last data we have for a domain with at least I<$MINVALS>
probes per server, rounded down to the hour

=over

=item I<$domain>

Name of the domain that is being monitored

=item I<$af>

Address family for domain servers used, either AF_INET (default) or AF_INET6

=back

The global default $<MINVALS> hold the minimum amount of probes that have valid
data that is required to make a sensible plot

=head3 RETURN VALUES

Time of last data seen, in seconds since the epoch and rounded down to the hour.

=head3 NOTES

Rounding down to the hour might be too restrictive when frequent updates are desired.
This is controlled by I<$LASTTROUNDDOWN> which could be used by applications as a policy tool.

=cut

sub last_val_domain {
    my($domain,$af) = @_;
    my($lastt) = 0;
    my($t);

    foreach my $s (domain_servers($domain,$af)) {
        $t = last_val($s,$af);
        $lastt = $t if $lastt==0 || $t < $lastt;
    }
    return($lastt)
}
        

=item ser_alias

resolve server aliases

=cut

sub ser_alias {
       my($s) = @_;
       my $seralias = $dns_config->get_seralias();
       return($seralias->{$s}) if ($seralias->{$s});
       return($s);
}


=item ser_name

return a readable server name

=cut

sub ser_name {
    my($s) = @_;
    my $sername = $dns_config->get_sernames(); 	
    return($sername->{$s}) if ($sername->{$s});
    return($s);
}
        

=item mon_alias

resolve probe alias

=cut

sub mon_alias {
    my($s) = @_;
    my $monalias = $dns_config->get_monalias();
    return($monalias->{$s}) if ($monalias->{$s});
    return($s);
}


=item mon_name

return a readable probe name

=cut

sub mon_name {
    my($s) = @_;
    my $monname = $dns_config->get_monnames();
    return($monname->{$s}) if ($monname->{$s});
    return($1) if ($s =~ /(.*)\.example\.net$/);
    return($s);
}

=item mk_pngpath

Make sure the path for the png file exists
a no-op if the flat file structure is used

=cut

sub mk_pngpath {
	my ($file) = @_;
	if (!$opt_flatpngdir)
	{
		my $old_umask = umask(0);
		eval { mkpath(dirname("$PNGDIR/$file"),0,0777) };
		if ($@) {
			$log->fatal("Couldn't create ", dirname("$PNGDIR/$file"), ": $@");
         }
		umask($old_umask);
	}
}

=item mk_pngname

construct a hopefully safe and unique graphics filename
room for imporovement here
      what parms need to go in?
      structure in subdirs to make directories smaller

=cut

sub mk_pngname {
    my($type, $start, $stop, $entity, %a) = @_;
    my(@a) = ();
    my(@e) = ();
    my $f = '';

    # the elements depend on the typ, cannot help it
    if ($type =~ /^S/) {
        @a = ('cf', 'maxms', 'valid', 'opt_w', 'opt_h');
        %a = (%DEFS_SRV_PLOT, %a);
    }
    elsif ($type =~ /^MD/) {
        @a = ('cf', 'maxpct', 'valid', 'opt_w', 'opt_h');
        %a = (%DEFS_MON_PLOT, %a);
    }
    elsif ($type =~ /^M/) {
        @a = ('cf', 'maxms', 'valid', 'opt_w', 'opt_h');
        %a = (%DEFS_MON_PLOT, %a);
    }
    elsif ($type =~ /^D/) {
        @a = ('cf', 'maxpct', 'valid', 'opt_w', 'opt_h');
        %a = (%DEFS_DOM_PLOT, %a);
    }
    else {
        $log->logdie("mk_pkgname: unknown PNG type '$type'");
    }

    # collect the elements
    push(@e, $type);
    push(@e, $entity);
    push(@e, "v6" ) if ($a{af} == &AF_INET6);
    push(@e, $start->strftime('%Y%m%d%H%M'));
    push(@e, $stop->strftime('%Y%m%d%H%M'));
	if (!$opt_flatpngdir)
	{
		$log->debug("elements: @e @a") if $opt_d;
		$f = join('/', map{ defined($_) && $_ ne '' ? $_ : "__DEFAULT__" } @e,
						join('_', map {defined($a{$_}) && $a{$_} ne '' ? $a{$_} : "__DEFAULT__"} @a));

		if ($opt_flatpngdir_transition_kludge)
		{
			# XXX transition hack, remove after 2 weeks when activated
			my $legacy_png_filename;
			$opt_flatpngdir = 1;
			$legacy_png_filename = mk_pngname(@_);
			$opt_flatpngdir = 0;

			if ( -r sprintf("%s/%s",get_pngdir(),$legacy_png_filename))	
			{
				$f = $legacy_png_filename;
			}
		}
		$log->debug($f) if $opt_d;
	}
	else
	{
		foreach my $a (@a) {
			push(@e, $a{$a});
		}
		$log->debug("elements: @e") if $opt_d;

		# make the name
		foreach $_ (@e) {
			next unless defined($_);
			s/\///g;
			s/ //g;
			s/_//g;
			$f .= "$_" . '_';
		}
	}
	
	$log->debug("generated pngfilename: \"${f}.png\"");
	
    chop($f); 
    return("${f}.png");    
}


=item plot_server

=head3 NAME

plot_server -- make a server plot

=head3 SYNOPSIS

plot_server(I<%a>)

=head3 DESCRIPTION

make a server plot; This is where it all happens!
arguments are in hash !

=over

=item I<cf>
consolidation function (Default: 'AVERAGE')

=item I<maxms>
y axis max value (Default: 250)

=item I<valid>
maximum percentage of invalid data to include a probe (Default: 10)

=item I<minmons>
minimum number of probes for a plot (Default: 5)

=item I<orange> 
orange color range (Default: '66:90')

=item I<red>
red color minimum (Default: '90')

=item I<opt_w>
plot width in pixels (Default: 800)

=item I<opt_h>
plot height in pixels (Default: 500)

=item I<opt_v>
verbose switch (Default: 0)

=item I<fuzzy>
fuzzy matching (Default: 0)

=item I<af>

address family, either AF_INET(default) or AF_INET6

=item I<force>

Create a plot, even when there is a cached one

=back

=head3 RETURN VALUES

return value is string with first letter coding the result

=over

=item B<F>
png file name (normal, we generated the file)

=item B<C>
png file name (normal, we found the cached file)

=item B<P>
png file name (fuzzily matched result, see below)

=item B<M>
User Digestable Diagnostic Message

=item B<E>
Error (Fatal for Admin Digestion)

=back

=cut

sub plot_server {
    my $func = 'DNSRRD::plot_server:';
    my $startTime=[gettimeofday];

    # argument processing
    my %a = (%DEFS_SRV_PLOT, @_); # argment hash values overwrite defaults
    if ($opt_d) {
        foreach my $k (sort keys %a) {
            $log->debug("$k = $a{$k}");
        }
    }

    # we need at least these
    return("E $func no 'start' argument") unless $a{start};
    return("E $func no 'stop' argument") unless $a{stop};
    return("E $func no 'srv' argument") unless $a{srv};

    # localise arguments
    my $minmons = $a{minmons};
    my $fuzzy = $a{fuzzy};
    my $opt_v = $a{opt_v};
    my @orange = split(/:/, $a{orange});
    my $red = $a{red};
    my $scale = 1000 / $a{maxms};
    my $miny = -0.1;
    my $srv = $a{srv};
    my $start = $a{start};
    my $starts = $start->strftime($RRDTOOL_AT_TIME_FMT);
    my $stop = $a{stop};
    my $stops = $stop->strftime($RRDTOOL_AT_TIME_FMT);
    my $valid = $a{valid};
	my $af    = $a{af};
	my $rrddir = $af == &AF_INET6 ? $RRDDIR6 : $RRDDIR;
	my $ipv = $af == &AF_INET ? "IPv4" : $af == &AF_INET6 ? "IPv6" : "Unknown";

    my $file = mk_pngname('S', $start, $stop, $srv, %a);

	if (not exists($a{refresh})) {
		return("C $PNGPATH/$file") if (-r "$PNGDIR/$file");    # return if it is already cached
	} else {
		unlink("$PNGDIR/$file");
		$log->info("Refreshing $PNGDIR/$file");
	}

    if ($fuzzy) {                       # accept the previous hour or previous 
                                        # day if available. this is used in 
                                        # overview pages to speed up the 
                                        # "user experience"
        my $offset = 0;


        if (($stop-$start) <= 48*60*60) {
            $offset = 60*60;
        }
        else {
            $offset = 24*60*60;
        }

        my $fstart = $start - $offset;
        my $fstop = $stop - $offset;
        my $fuzzyfile = mk_pngname('S', $fstart, $fstop, $srv, %a);

        # fuzzyily cached
	$log->warn("fuzzy mismatch on $PNGPATH/$fuzzyfile, file not found!") if (! -r "$PNGDIR/$fuzzyfile");
        return("P $PNGPATH/$fuzzyfile") if (-r "$PNGDIR/$fuzzyfile"); 
    }

	mk_pngpath($file);
    $locker->lock("$PNGDIR/$file") || $log->logdie("$func: cannot lock $PNGDIR/$file");

    if (-r "$PNGDIR/$file") {    # somebody else made it while we were waiting
        $log->debug("$file was generated while we slept delta=", tv_interval($startTime));
        $locker->unlock("$PNGDIR/$file") || $log->logdie("$func: cannot unlock $PNGDIR/$file");
        return("C $PNGPATH/$file");
    }
    
    $log->debug("Generating new image: $file");
    
    ### OK no cached graphics, fuzzy or not; we have to make it ourselves.
    #
    my @mons = sort (select_mons($starts, $stops, $valid, $srv, $af, &RIPE::DNSMon::RRD::all_mons($af)));
    if ($#mons <= $minmons) {
        my $msg = "<h2>";
        $msg .= $#mons<0? "No" : sprintf "Only %d", $#mons+1;
        $msg .= " probes with sufficient valid $ipv data <br>";
        $msg .= "for domain server $srv<br>"; 
        $msg .= "and period $starts to $stops</h2>";
        symlink($NOGRAPH, "$PNGDIR/$file") || $log->logdie("$func: cannot symlink($NOGRAPH, $PNGDIR/$file)");
        $locker->unlock("$PNGDIR/$file") || $log->logdie("$func: cannot unlock $PNGDIR/$file");
        return("M $msg");
    }

    # make short names for RRD variables
    my %mnam = &mknames(@mons);

    # assemble arguments for RRD::graph function
    my(@go) = ();     
    push @go, "$PNGDIR/$file";
    push @go, '-a', 'PNG';
    push @go, '-s', $starts;
    push @go, '-e', $stops;
    push @go, '-w',  $a{opt_w};
    push @go, '-h',  $a{opt_h};
    push @go, '-l', sprintf '%f', $miny;
    push @go, '-u', sprintf '%f', ($#mons+1);
    push @go, '--rigid';
    push @go, '--title', sprintf "$ipv Query Delays ($a{cf})  for  %s  [$starts - $stops UTC]", &ser_name($srv);
    push @go, '-v', "$a{cf} of Response Time [0-$a{maxms}] (ms)";

    foreach my $mon (@mons) {           # you need to really grok RRD::graph to
                                        # understand this, use opt_d
        # define vars from RRD
        push @go, "DEF:$mnam{$mon}_VAL=$rrddir/$mon/$srv:RTT:$a{cf}";
        push @go, "DEF:$mnam{$mon}_DRP=$rrddir/$mon/$srv:DROPS:MAX";

        # SCL - scale it down to [0:1]
        push @go, "CDEF:$mnam{$mon}_SCL=$mnam{$mon}_VAL,$scale,*";
        # CUT - cut off at 1 (paranoid)
        push @go, "CDEF:$mnam{$mon}_CUT=$mnam{$mon}_SCL,1,GT,1,$mnam{$mon}_SCL,IF"; 
        # VAR defined if drop rate is below the orange range
        push @go, "CDEF:$mnam{$mon}_VAR=$mnam{$mon}_DRP,$orange[0],GT,UNKN,$mnam{$mon}_CUT,IF";
        # ORA - defined dif drop rate is within orange range
        push @go, "CDEF:$mnam{$mon}_ORA=$mnam{$mon}_DRP,$orange[0],GT,$mnam{$mon}_DRP,$orange[1],LT,+,2,EQ,$mnam{$mon}_CUT,UNKN,IF";
        # RED - defined if drop rate in red range
        push @go, "CDEF:$mnam{$mon}_RED=$mnam{$mon}_DRP,$red,LT,UNKN,$mnam{$mon}_CUT,IF";
        push @go, "CDEF:$mnam{$mon}_MAX=1,1,$mnam{$mon}_CUT,-,MIN";
        push @go, "CDEF:$mnam{$mon}_ALL=$mnam{$mon}_DRP,UN,$mnam{$mon}_CUT,UN,+,1,EQ,1,0,IF";
        push @go, "CDEF:$mnam{$mon}_UND=$mnam{$mon}_DRP,UN,1,0,IF";
    } 
    my $stack = 0;
	# XXX  FIXME stack usage not compatible with rrdtool 1.2.x!
    foreach my $mon (@mons) {
        if ($stack == 0) {     # bottom most (start with AREA)
            push @go, sprintf "AREA:$mnam{$mon}_VAR#$BLUE:%02d\\:%-11.11s", $stack, &mon_name($mon);
            push @go, "STACK:$mnam{$mon}_ORA#$ORANGE:";
            push @go, "STACK:$mnam{$mon}_RED#$RED:";
            push @go, "STACK:$mnam{$mon}_ALL#$RED:";
            push @go, "STACK:$mnam{$mon}_MAX#$WHITE:";
            push @go, "STACK:$mnam{$mon}_UND#$BLACK:";
        }
        elsif ($stack % 2) {    # even numbered are light color
            push @go, sprintf "STACK:$mnam{$mon}_VAR#$LGTBLUE:%02d\\:%-11.11s", $stack, &mon_name($mon);
            push @go, "STACK:$mnam{$mon}_ORA#$ORANGE:";
            push @go, "STACK:$mnam{$mon}_RED#$RED:";
            push @go, "STACK:$mnam{$mon}_ALL#$RED:";
            push @go, "STACK:$mnam{$mon}_MAX#$WHITE:";
            push @go, "STACK:$mnam{$mon}_UND#$BLACK:";
        }
        else {            # odd numbered are dark color
            push @go, sprintf "STACK:$mnam{$mon}_VAR#$BLUE:%02d\\:%-11.11s", $stack, &mon_name($mon);
            push @go, "STACK:$mnam{$mon}_ORA#$ORANGE:";
            push @go, "STACK:$mnam{$mon}_RED#$RED:";
            push @go, "STACK:$mnam{$mon}_ALL#$RED:";
            push @go, "STACK:$mnam{$mon}_MAX#$WHITE:";
            push @go, "STACK:$mnam{$mon}_UND#$BLACK:";
        }
        $stack++;
    }

    # the use of HRULE below is devious, it prints one over the other
    # we do this for the side effect in the legend.
    # make sure the last is black!
    # also make sure the yvalue is in the plot and on a grid line
    #
    while ($stack%8) {        # pad out the legend
        push @go, "HRULE:$miny#$WHITE:              ";
        $stack++;
    }
    push @go, "HRULE:$miny#$ORANGE:$orange[0]% < MAX(Unanswered Queries) < $orange[1]%  ";
    push @go, "HRULE:$miny#$RED:MAX(Unanswered Queries) >= $red%  ";
    push @go, "HRULE:$miny#$BLACK:no data available  ";
    $log->debug("graph(@go)") if ($log->is_debug && $opt_d);

    # this is where it gets done!
    my($gret, $x, $y) = RRDs::graph @go;
    if (my $ERROR = RRDs::error) {
		$log->error("RRDs::error returned: \"$ERROR\"");
        $locker->unlock("$PNGDIR/$file") || $log->logdie("$func: cannot unlock $PNGDIR/$file");
        return("E $func: RRDs::graph ERROR: $ERROR\n");
    }
    $log->debug(": $x x $y ", (join "/", @$gret)) if $opt_v; 
    
    $log->debug("$file generated, delta=",tv_interval($startTime));
    
    chmod(0664, "$PNGDIR/$file") || $log->logdie("chmod(0664, $PNGDIR/$file)");
    $locker->unlock("$PNGDIR/$file") || $log->logdie("$func: cannot unlock $PNGDIR/$file");
    return("F $PNGPATH/$file");
}


=item plot_mon

=head3 NAME

plot_mon -- make a monitor/probe plot

=head3 SYNOPSIS

plot_mon(I<%a>)

=head3 DESCRIPTION

make a monitor/probe plot
arguments are in hash !

=over

=item I<cf>
consolidation function (Default: 'AVERAGE')

=item I<maxms>
y axis max value (Default: 250)

=item I<valid>
maximum percentage of invalid data to include a probe (Default: 10)

=item I<orange>
orange color range (Default: '66:90')

=item I<red>
red color minimum (Default: '90')

=item I<opt_w>
plot width in pixels (Default: 800)

=item I<opt_h>
plot height in pixels (Default: 500)

=item I<opt_v>
verbose switch (Default: 0)

=item I<fuzzy>
fuzzy matching (Default: 0)

=item I<af>

address family, either AF_INET(default) or AF_INET6

=item I<force>

Create a plot, even when there is a cached one

=back

=head3 RETURN VALUES

return value is string with first letter coding the result

=head3 SEE ALSO

see L</plot_server> for explanations

=cut

sub plot_mon {

    my $func = 'DNSRRD::plot_mon:';

    # argument processing
    my %a = (%DEFS_MON_PLOT, @_); # argment hash values overwrite defaults
    if ($opt_d) {
        foreach my $k (sort keys %a) {
            $log->debug("$k = $a{$k}");
        }
    }

    # we need at least these
    return("E $func no 'start' argument") unless $a{start};
    return("E $func no 'stop' argument") unless $a{stop};
    return("E $func no 'mon' argument") unless $a{mon};

    # localise arguments
    my $mon = $a{mon};
    my $fuzzy = $a{fuzzy};
    my $opt_v = $a{opt_v};
    my @orange = split(/:/, $a{orange});
    my $red = $a{red};
    my $scale = 1000 / $a{maxms};
    my $miny = -0.1;
    my $srv = $a{srv};
    my $start = $a{start};
    my $starts = $start->strftime($RRDTOOL_AT_TIME_FMT);
    my $stop = $a{stop};
    my $stops = $stop->strftime($RRDTOOL_AT_TIME_FMT);
    my $valid = $a{valid};
	my $af    = $a{af};
	my $rrddir = $af == &AF_INET6 ? $RRDDIR6 : $RRDDIR;
	my $ipv = $af == &AF_INET ? "IPv4" : $af == &AF_INET6 ? "IPv6" : "Unknown";

    my $file = mk_pngname('M', $start, $stop, $mon, %a);

	if (not exists($a{refresh})) {
		return("C $PNGPATH/$file") if (-r "$PNGDIR/$file"); # cached
	} else {
		unlink("$PNGDIR/$file");
		$log->info("Refreshing $PNGDIR/$file");
	}

    if ($fuzzy) {
        my $offset = 0;

        if (($stop-$start) <= 48*60*60) {
            $offset = 60*60;
        }
        else {
            $offset = 24*60*60;
        }

        my $fstart = $start - $offset;
        my $fstop = $stop - $offset;
        my $fuzzyfile = mk_pngname('M', $fstart, $fstop, $mon, %a);

	$log->warn("fuzzy mismatch on $PNGPATH/$fuzzyfile, file not found!") if (! -r "$PNGDIR/$fuzzyfile");
        return("P $PNGPATH/$fuzzyfile") if (-r "$PNGDIR/$fuzzyfile"); # fuzzyily cached
    }

	mk_pngpath($file);
    $locker->lock("$PNGDIR/$file") || $log->logdie("$func: cannot lock $PNGDIR/$file");

    if (-r "$PNGDIR/$file") {    # somebody else made it while we were waiting
        $locker->unlock("$PNGDIR/$file") || $log->logdie("$func: cannot unlock $PNGDIR/$file");
        return("C $PNGPATH/$file");
    }

      my @servers = sort (select_servers($starts, $stops, $valid, $mon, $af, &RIPE::DNSMon::RRD::mon_servers($mon,$af)));
      if ($#servers <= 0) {
        my $msg = "<h2>";
        $msg .= "No valid $ipv data <br>";
        $msg .= "for probe $mon<br>"; 
        $msg .= "and period $starts to $stops</h2>";
        symlink($NOGRAPH, "$PNGDIR/$file") || $log->logdie("$func: cannot symlink($NOGRAPH, $PNGDIR/$file)");
        $locker->unlock("$PNGDIR/$file") || $log->logdie("$func: cannot unlock $PNGDIR/$file");
        return("M $msg");
    }

    # make short names
    my %mnam = &mknames(@servers);

    # assemble plot arguments
    my(@go) = ();     # RRD::graph options
    push @go, "$PNGDIR/$file";
    push @go, '-a', 'PNG';
    push @go, '-s', $starts;
    push @go, '-e', $stops;
    push @go, '-w',  $a{opt_w};
    push @go, '-h',  $a{opt_h};
    push @go, '-l', sprintf '%f', $miny;
    push @go, '-u', sprintf '%f', ($#servers+1);
    push @go, '--rigid';
    push @go, '--title', sprintf "$ipv Query Delays ($a{cf})  measured at %s  [$starts - $stops UTC]", $mon;
    push @go, '-v', "$a{cf} of Response Time [0-$a{maxms}] (ms)";

    foreach my $srv (@servers) {
        # define vars from RRD
        push @go, "DEF:$mnam{$srv}_VAL=$rrddir/$mon/$srv:RTT:$a{cf}";
        push @go, "DEF:$mnam{$srv}_DRP=$rrddir/$mon/$srv:DROPS:MAX";

        # SCL - scale it down to [0:1]
        push @go, "CDEF:$mnam{$srv}_SCL=$mnam{$srv}_VAL,$scale,*";
        # CUT - cut off at 1 (paranoid)
        push @go, "CDEF:$mnam{$srv}_CUT=$mnam{$srv}_SCL,1,GT,1,$mnam{$srv}_SCL,IF"; 
        # VAR defined if drop rate is below the orange range
        push @go, "CDEF:$mnam{$srv}_VAR=$mnam{$srv}_DRP,$orange[0],GT,UNKN,$mnam{$srv}_CUT,IF";
        # ORA - defined dif drop rate is within orange range
        push @go, "CDEF:$mnam{$srv}_ORA=$mnam{$srv}_DRP,$orange[0],GT,$mnam{$srv}_DRP,$orange[1],LT,+,2,EQ,$mnam{$srv}_CUT,UNKN,IF";
        # RED - defined if drop rate in red range
        push @go, "CDEF:$mnam{$srv}_RED=$mnam{$srv}_DRP,$red,LT,UNKN,$mnam{$srv}_CUT,IF";
        push @go, "CDEF:$mnam{$srv}_MAX=1,1,$mnam{$srv}_CUT,-,MIN";
        push @go, "CDEF:$mnam{$srv}_ALL=$mnam{$srv}_DRP,UN,$mnam{$srv}_CUT,UN,+,1,EQ,1,0,IF";
        push @go, "CDEF:$mnam{$srv}_UND=$mnam{$srv}_DRP,UN,1,0,IF";
    } 
    my $stack = 0;
	# XXX  FIXME stack usage not compatible with rrdtool 1.2.x!
    foreach my $srv (@servers) {
        if ($stack == 0) {
            push @go, sprintf "AREA:$mnam{$srv}_VAR#$BLUE:%02d\\: %-25.25s", $stack, $srv;
            push @go, "STACK:$mnam{$srv}_ORA#$ORANGE:";
            push @go, "STACK:$mnam{$srv}_RED#$RED:";
            push @go, "STACK:$mnam{$srv}_ALL#$RED:";
            push @go, "STACK:$mnam{$srv}_MAX#$WHITE:";
            push @go, "STACK:$mnam{$srv}_UND#$BLACK:";
        }
        elsif ($stack % 2) {
            push @go, sprintf "STACK:$mnam{$srv}_VAR#$LGTBLUE:%02d\\: %-25.25s", $stack, $srv;
            push @go, "STACK:$mnam{$srv}_ORA#$ORANGE:";
            push @go, "STACK:$mnam{$srv}_RED#$RED:";
            push @go, "STACK:$mnam{$srv}_ALL#$RED:";
            push @go, "STACK:$mnam{$srv}_MAX#$WHITE:";
            push @go, "STACK:$mnam{$srv}_UND#$BLACK:";
        }
        else {
            push @go, sprintf "STACK:$mnam{$srv}_VAR#$BLUE:%02d\\: %-25.25s", $stack, $srv;
            push @go, "STACK:$mnam{$srv}_ORA#$ORANGE:";
            push @go, "STACK:$mnam{$srv}_RED#$RED:";
            push @go, "STACK:$mnam{$srv}_ALL#$RED:";
            push @go, "STACK:$mnam{$srv}_MAX#$WHITE:";
            push @go, "STACK:$mnam{$srv}_UND#$BLACK:";
        }
        $stack++;
    }

    while ($stack%4) {
        push @go, "HRULE:$miny#$WHITE:                             ";
        $stack++;
    }
    push @go, "HRULE:$miny#$ORANGE:$orange[0]% < MAX(Unanswered Queries) < $orange[1]%  ";
    push @go, "HRULE:$miny#$RED:MAX(Unanswered Queries) >= $red%  ";
    push @go, "HRULE:$miny#$BLACK:no data available  ";

    $log->debug("graph(@go)") if ($log->is_debug && $opt_d);
    my($gret, $x, $y) = RRDs::graph @go;
    if (my $ERROR = RRDs::error) {
		$log->error("RRDs::error returned: \"$ERROR\"");
        $locker->unlock("$PNGDIR/$file") || $log->logdie("$func: cannot unlock $PNGDIR/$file");
        return("E $func: RRDs::graph ERROR: $ERROR\n");
    }
    $log->debug(": $x x $y ", (join "/", @$gret)) if $opt_v; 

    chmod(0664, "$PNGDIR/$file") || $log->logdie("chmod(0664, $PNGDIR/$file)");
    $locker->unlock("$PNGDIR/$file") || $log->logdie("$func: cannot unlock $PNGDIR/$file");
    return("F $PNGPATH/$file");
}


=item plot_server_drops

=head3 NAME

plot_server_drops -- make a server plot with unanswered queries

=head3 SYNOPSIS

plot_server_drops(I<%a>)

=head3 DESCRIPTION

make a server plot with unanswered queries
arguments are in hash !

=over

=item I<cf>
consolidation function (Default: 'AVERAGE')

=item I<maxpct>
y axis max value (Default: 33)

=item I<valid>
maximum percentage of invalid data to include a probe (Default: 10)

=item I<minmons>
minimum number of probes for a plot (Default: 5)

=item I<orange>
orange color range (Default: '66:90')

=item I<red>
red color minimum (Default: '90')

=item I<opt_w>
plot width in pixels (Default: 800)

=item I<opt_h>
plot height in pixels (Default: 500)

=item I<opt_v>
verbose switch (Default: 0)

=item I<fuzzy>
fuzzy matching (Default: 0)

=item I<af>

address family, either AF_INET(default) or AF_INET6

=item I<force>

Create a plot, even when there is a cached one

=back


=head3 RETURN VALUES

return value is string with first letter coding the result
see plot_server for explanations

=head3 SEE ALSO

see L</plot_server> for explanations

=cut

sub plot_server_drops {
    my $func = 'DNSRRD::plot_server_drops:';

    # argument processing
    my %a = (%DEFS_SRV_PLOT_DROPS, @_); # argment hash values overwrite defaults
    if ($opt_d) {
        foreach my $k (sort keys %a) {
            $log->debug("$k = $a{$k}");
        }
    }

    # we need at least these
    return("E $func no 'start' argument") unless $a{start};
    return("E $func no 'stop' argument") unless $a{stop};
    return("E $func no 'srv' argument") unless $a{srv};

    # localise arguments
    my $maxpct = $a{maxpct};
    my $minmons = $a{minmons};
    my $miny = -0.1;
    my $opt_v = $a{opt_v};
    my $fuzzy = $a{fuzzy};
    my @orange = split(/:/, $a{orange});
    my $red = $a{red};
    my $srv = $a{srv};
    my $start = $a{start};
    my $starts = $start->strftime($RRDTOOL_AT_TIME_FMT);
    my $stop = $a{stop};
    my $stops = $stop->strftime($RRDTOOL_AT_TIME_FMT);
    my $valid = $a{valid};
	my $af    = $a{af};
	my $rrddir = $af == &AF_INET6 ? $RRDDIR6 : $RRDDIR;
	my $ipv = $af == &AF_INET ? "IPv4" : $af == &AF_INET6 ? "IPv6" : "Unknown";

    $maxpct = 50 unless $maxpct;
    $maxpct = 100 if $maxpct > 100;
    my $scale = 1 / $maxpct;

    my $file = mk_pngname('SD', $start, $stop, $srv, %a);

	if (not exists($a{refresh})) {
		return("C $PNGPATH/$file") if (-r "$PNGDIR/$file"); # cached
	} else {
		unlink("$PNGDIR/$file");
		$log->info("Refreshing $PNGDIR/$file");
	}

    if ($fuzzy) {
        my $offset = 0;

        if (($stop-$start) <= 48*60*60) {
            $offset = 60*60;
        }
        else {
            $offset = 24*60*60;
        }

        my $fstart = $start - $offset;
        my $fstop = $stop - $offset;
        my $fuzzyfile = mk_pngname('SD', $fstart, $fstop, $srv, %a);

	$log->warn("fuzzy mismatch on $PNGPATH/$fuzzyfile, file not found!") if (! -r "$PNGDIR/$fuzzyfile");
        return("P $PNGPATH/$fuzzyfile") if (-r "$PNGDIR/$fuzzyfile"); # fuzzyily cached
    }

	mk_pngpath($file);
    $locker->lock("$PNGDIR/$file") || $log->logdie("$func: cannot lock $PNGDIR/$file");

    if (-r "$PNGDIR/$file") {    # somebody else made it while we were waiting
        $locker->unlock("$PNGDIR/$file") || $log->logdie("$func: cannot unlock $PNGDIR/$file");
        return("C $PNGPATH/$file");
    }

    my @mons = sort (select_mons($starts, $stops, $valid, $srv, $af, &RIPE::DNSMon::RRD::all_mons($af)));
    if ($#mons <= $minmons) {
        my $msg = "<h2>";
        $msg .= $#mons<0? "No" : sprintf "Only %d", $#mons+1;
        $msg .= " probes with sufficient valid $ipv data <br>";
        $msg .= "for domain server $srv<br>"; 
        $msg .= "and period $starts to $stops</h2>";
        symlink($NOGRAPH, "$PNGDIR/$file") || $log->logdie("$func: cannot symlink($NOGRAPH, $PNGDIR/$file)");
        $locker->unlock("$PNGDIR/$file") || $log->logdie("$func: cannot unlock $PNGDIR/$file");
        return("M $msg");
    }

    # make short names
    my %mnam = &mknames(@mons);

    # assemble plot arguments
    my(@go) = ();     # RRD::graph options
    push @go, "$PNGDIR/$file";
    push @go, '-a', 'PNG';
    push @go, '-s', $starts;
    push @go, '-e', $stops;
    push @go, '-w',  $a{opt_w};
    push @go, '-h',  $a{opt_h};
    push @go, '-l', sprintf '%f', $miny;
    push @go, '-u', sprintf '%f', ($#mons+1);
    push @go, '--rigid';
    push @go, '--title', sprintf "$ipv Unanswered Queries ($a{cf})  for  %s  [$starts - $stops UTC]", &ser_name($srv);
    push @go, '-v', "$a{cf} of Unanswered Queries [0-$maxpct] (%)";

    foreach my $mon (@mons) {
        # define vars from RRD
        push @go, "DEF:$mnam{$mon}_VAL=$rrddir/$mon/$srv:DROPS:$a{cf}";
        push @go, "DEF:$mnam{$mon}_DRP=$rrddir/$mon/$srv:DROPS:MAX";

        # SCL - scale it down to [0:1]
        push @go, "CDEF:$mnam{$mon}_SCL=$mnam{$mon}_VAL,$scale,*";
        # CUT - cut off at 1 (paranoid)
        push @go, "CDEF:$mnam{$mon}_CUT=$mnam{$mon}_SCL,1,GT,1,$mnam{$mon}_SCL,IF"; 
        # VAR defined if drop rate is below the orange range
        push @go, "CDEF:$mnam{$mon}_VAR=$mnam{$mon}_DRP,$orange[0],GT,UNKN,$mnam{$mon}_CUT,IF";
        # ORA - defined dif drop rate is within orange range
        push @go, "CDEF:$mnam{$mon}_ORA=$mnam{$mon}_DRP,$orange[0],GT,$mnam{$mon}_DRP,$orange[1],LT,+,2,EQ,$mnam{$mon}_CUT,UNKN,IF";
        # RED - defined if drop rate in red range
        push @go, "CDEF:$mnam{$mon}_RED=$mnam{$mon}_DRP,$red,LT,UNKN,$mnam{$mon}_CUT,IF";
        push @go, "CDEF:$mnam{$mon}_MAX=1,1,$mnam{$mon}_CUT,-,MIN";
        push @go, "CDEF:$mnam{$mon}_ALL=$mnam{$mon}_DRP,UN,$mnam{$mon}_CUT,UN,+,1,EQ,1,0,IF";
        push @go, "CDEF:$mnam{$mon}_UND=$mnam{$mon}_DRP,UN,1,0,IF";
    } 
    my $stack = 0;
	# XXX  FIXME stack usage not compatible with rrdtool 1.2.x!
    foreach my $mon (@mons) {
        if ($stack == 0) {
            push @go, sprintf "AREA:$mnam{$mon}_VAR#$YELLOW:%02d\\:%-11.11s", $stack, &mon_name($mon);
            push @go, "STACK:$mnam{$mon}_ORA#$ORANGE:";
            push @go, "STACK:$mnam{$mon}_RED#$RED:";
            push @go, "STACK:$mnam{$mon}_ALL#$RED:";
            push @go, "STACK:$mnam{$mon}_MAX#$GREEN:";
            push @go, "STACK:$mnam{$mon}_UND#$WHITE:";
        }
        else {
            push @go, sprintf "STACK:$mnam{$mon}_VAR#$YELLOW:%02d\\:%-11.11s", $stack, &mon_name($mon);
            push @go, "STACK:$mnam{$mon}_ORA#$ORANGE:";
            push @go, "STACK:$mnam{$mon}_RED#$RED:";
            push @go, "STACK:$mnam{$mon}_ALL#$RED:";
            push @go, "STACK:$mnam{$mon}_MAX#$GREEN:";
            push @go, "STACK:$mnam{$mon}_UND#$WHITE:";
        }
        $stack++;
    }

    while ($stack%8) {
        push @go, "HRULE:$miny#$WHITE:              ";
        $stack++;
    }
    push @go, "HRULE:$miny#$ORANGE:$orange[0]% < MAX(Unanswered Queries) < $orange[1]%  ";
    push @go, "HRULE:$miny#$RED:MAX(Unanswered Queries) >= $red%  ";
    push @go, "HRULE:$miny#$WHITE:no data available  ";

    $log->debug("graph(@go)") if ($log->is_debug && $opt_d);
    my($gret, $x, $y) = RRDs::graph @go;
    if (my $ERROR = RRDs::error) {
		$log->error("RRDs::error returned: \"$ERROR\"");
        $locker->unlock("$PNGDIR/$file") || $log->logdie("$func: cannot unlock $PNGDIR/$file");
        return("E $func: RRDs::graph ERROR: $ERROR\n");
    }
    $log->debug(": $x x $y ", (join "/", @$gret)) if $opt_v; 

    chmod(0664, "$PNGDIR/$file") || $log->logdie("chmod(0664, $PNGDIR/$file)");
    $locker->unlock("$PNGDIR/$file") || $log->logdie("$func: cannot unlock $PNGDIR/$file");
    return("F $PNGPATH/$file");
}



=item plot_mon_drops

=head3 NAME

make a server plot with unanswered queries

=head3 SYNOPSIS

plot_mon_drops(I<%a>)

=head3 DESCRIPTION

make a server plot with unanswered queries
arguments are in hash !

=over

=item I<cf>
consolidation function (Default: 'AVERAGE')

=item I<maxpct>
y axis max value (Default: 33)

=item I<valid>
maximum percentage of invalid data to include a probe (Default: 10)

=item I<orange>
orange color range (Default: '66:90')

=item I<red>
red color minimum (Default: '90')

=item I<opt_w>
plot width in pixels (Default: 800)

=item I<opt_h>
plot height in pixels (Default: 500)

=item I<opt_v>
verbose switch (Default: 0)

=item I<fuzzy>
fuzzy matching (Default: 0)

=item I<af>

address family, either AF_INET(default) or AF_INET6

=item I<force>

Create a plot, even when there is a cached one

=back

=head3 RETURN VALUES

return value is string with first letter coding the result

=head3 SEE ALSO

see L</plot_server> for explanations

=cut

sub plot_mon_drops {
    my $func = 'DNSRRD::plot_mon:';

    # argument processing
    my %a = (%DEFS_MON_PLOT_DROPS, @_); # argment hash values overwrite defaults
    if ($opt_d) {
        foreach my $k (sort keys %a) {
            $log->debug("$k = $a{$k}");
        }
    }

    # we need at least these
    return("E $func no 'start' argument") unless $a{start};
    return("E $func no 'stop' argument") unless $a{stop};
    return("E $func no 'mon' argument") unless $a{mon};

    # localise arguments
    my $mon = $a{mon};
    my $maxpct = $a{maxpct};
    my $minmons = $a{minmons};
    my $miny = -0.1;
    my $opt_v = $a{opt_v};
    my $fuzzy = $a{fuzzy};
    my @orange = split(/:/, $a{orange});
    my $red = $a{red};
    my $srv = $a{srv};
    my $start = $a{start};
    my $starts = $start->strftime($RRDTOOL_AT_TIME_FMT);
    my $stop = $a{stop};
    my $stops = $stop->strftime($RRDTOOL_AT_TIME_FMT);
    my $valid = $a{valid};
	my $af    = $a{af};
	my $rrddir = $af == &AF_INET6 ? $RRDDIR6 : $RRDDIR;
	my $ipv = $af == &AF_INET ? "IPv4" : $af == &AF_INET6 ? "IPv6" : "Unknown";

    $maxpct = 50 unless $maxpct;
    $maxpct = 100 if $maxpct > 100;
    my $scale = 1 / $maxpct;

    my $file = mk_pngname('MD', $start, $stop, $mon, %a);

	if (not exists($a{refresh})) {
		return("C $PNGPATH/$file") if (-r "$PNGDIR/$file"); # cached
	} else {
		unlink("$PNGDIR/$file");
		$log->info("Refreshing $PNGDIR/$file");
	}

    if ($fuzzy) {
        my $offset = 0;

        if (($stop-$start) <= 48*60*60) {
            $offset = 60*60;
        }
        else {
            $offset = 24*60*60;
        }

        my $fstart = $start - $offset;
        my $fstop = $stop - $offset;
        my $fuzzyfile = mk_pngname('MD', $fstart, $fstop, $mon, %a);

	$log->warn("fuzzy mismatch on $PNGPATH/$fuzzyfile, file not found!") if (! -r "$PNGDIR/$fuzzyfile");
        return("P $PNGPATH/$fuzzyfile") if (-r "$PNGDIR/$fuzzyfile"); # fuzzyily cached
    }

	mk_pngpath($file);
    $locker->lock("$PNGDIR/$file") || $log->logdie("$func: cannot lock $PNGDIR/$file");

    if (-r "$PNGDIR/$file") {    # somebody else made it while we were waiting
        $locker->unlock("$PNGDIR/$file") || $log->logdie("$func: cannot unlock $PNGDIR/$file");
        return("C $PNGPATH/$file");
    }

      my @servers = sort (select_servers($starts, $stops, $valid, $mon, $af, &RIPE::DNSMon::RRD::mon_servers($mon,$af)));
      if ($#servers <= 0) {
        my $msg = "<h2>";
        $msg .= "No valid $ipv data <br>";
        $msg .= "for probe $mon<br>"; 
        $msg .= "and period $starts to $stops</h2>";
        symlink($NOGRAPH, "$PNGDIR/$file") || $log->logdie("$func: cannot symlink($NOGRAPH, $PNGDIR/$file)");
        $locker->unlock("$PNGDIR/$file") || $log->logdie("$func: cannot unlock $PNGDIR/$file");
        return("M $msg");
    }

    # make short names
    my %mnam = &mknames(@servers);

    # assemble plot arguments
    my(@go) = ();     # RRD::graph options
    push @go, "$PNGDIR/$file";
    push @go, '-a', 'PNG';
    push @go, '-s', $starts;
    push @go, '-e', $stops;
    push @go, '-w',  $a{opt_w};
    push @go, '-h',  $a{opt_h};
    push @go, '-l', sprintf '%f', $miny;
    push @go, '-u', sprintf '%f', ($#servers+1);
    push @go, '--rigid';
    push @go, '--title', sprintf "$ipv Unanswered Queries ($a{cf})  measured at %s  [$starts - $stops UTC]", $mon;
    push @go, '-v', "$a{cf} of Unanswered Queries [0-$maxpct] (%)";

    foreach my $srv (@servers) {
        # define vars from RRD
        push @go, "DEF:$mnam{$srv}_VAL=$rrddir/$mon/$srv:DROPS:$a{cf}";
        push @go, "DEF:$mnam{$srv}_DRP=$rrddir/$mon/$srv:DROPS:MAX";

        # SCL - scale it down to [0:1]
        push @go, "CDEF:$mnam{$srv}_SCL=$mnam{$srv}_VAL,$scale,*";
        # CUT - cut off at 1 (paranoid)
        push @go, "CDEF:$mnam{$srv}_CUT=$mnam{$srv}_SCL,1,GT,1,$mnam{$srv}_SCL,IF"; 
        # VAR defined if drop rate is below the orange range
        push @go, "CDEF:$mnam{$srv}_VAR=$mnam{$srv}_DRP,$orange[0],GT,UNKN,$mnam{$srv}_CUT,IF";
        # ORA - defined dif drop rate is within orange range
        push @go, "CDEF:$mnam{$srv}_ORA=$mnam{$srv}_DRP,$orange[0],GT,$mnam{$srv}_DRP,$orange[1],LT,+,2,EQ,$mnam{$srv}_CUT,UNKN,IF";
        # RED - defined if drop rate in red range
        push @go, "CDEF:$mnam{$srv}_RED=$mnam{$srv}_DRP,$red,LT,UNKN,$mnam{$srv}_CUT,IF";
        push @go, "CDEF:$mnam{$srv}_MAX=1,1,$mnam{$srv}_CUT,-,MIN";
        push @go, "CDEF:$mnam{$srv}_ALL=$mnam{$srv}_DRP,UN,$mnam{$srv}_CUT,UN,+,1,EQ,1,0,IF";
        push @go, "CDEF:$mnam{$srv}_UND=$mnam{$srv}_DRP,UN,1,0,IF";
    } 
    my $stack = 0;
	# XXX  FIXME stack usage not compatible with rrdtool 1.2.x!
    foreach my $srv (@servers) {
        if ($stack == 0) {
            push @go, sprintf "AREA:$mnam{$srv}_VAR#$YELLOW:%02d\\: %-25.25s", $stack, $srv;
            push @go, "STACK:$mnam{$srv}_ORA#$ORANGE:";
            push @go, "STACK:$mnam{$srv}_RED#$RED:";
            push @go, "STACK:$mnam{$srv}_ALL#$RED:";
            push @go, "STACK:$mnam{$srv}_MAX#$GREEN:";
            push @go, "STACK:$mnam{$srv}_UND#$WHITE:";
        }
        else {
            push @go, sprintf "STACK:$mnam{$srv}_VAR#$YELLOW:%02d\\: %-25.25s", $stack, $srv;
            push @go, "STACK:$mnam{$srv}_ORA#$ORANGE:";
            push @go, "STACK:$mnam{$srv}_RED#$RED:";
            push @go, "STACK:$mnam{$srv}_ALL#$RED:";
            push @go, "STACK:$mnam{$srv}_MAX#$GREEN:";
            push @go, "STACK:$mnam{$srv}_UND#$WHITE:";
        }
        $stack++;
    }

    while ($stack%4) {
        push @go, "HRULE:$miny#$WHITE:                             ";
        $stack++;
    }
    push @go, "HRULE:$miny#$ORANGE:$orange[0]% < MAX(Unanswered Queries) < $orange[1]%  ";
    push @go, "HRULE:$miny#$RED:MAX(Unanswered Queries) >= $red%  ";
    push @go, "HRULE:$miny#$WHITE:no data available  ";

    $log->debug("graph(@go)") if ($log->is_debug && $opt_d);
    my($gret, $x, $y) = RRDs::graph @go;
    if (my $ERROR = RRDs::error) {
		$log->error("RRDs::error returned: \"$ERROR\"");
        $locker->unlock("$PNGDIR/$file") || $log->logdie("$func: cannot unlock $PNGDIR/$file");
        return("E $func: RRDs::graph ERROR: $ERROR\n");
    }
    $log->debug(": $x x $y ", (join "/", @$gret)) if $opt_v; 

    chmod(0664, "$PNGDIR/$file") || $log->logdie("chmod(0664, $PNGDIR/$file)");
    $locker->unlock("$PNGDIR/$file") || $log->logdie("$func: cannot unlock $PNGDIR/$file");
    return("F $PNGPATH/$file");
}


=item plot_domain

=head3 NAME

plot_domain -- make a domain plot

=head3 SYNOPSIS

plot_domain(I<%a>)

=head3 DESCRIPTION

make a server plot
arguments are in hash !

=over

=item I<cf>
consolidation function (Default: 'AVERAGE')

=item I<maxms>
y axis max value (Default: 250)

=item I<valid>
maximum percentage of invalid data to include a probe (Default: 10)

=item I<orange>
orange color range (Default: '66:90')

=item I<red>
red color minimum (Default: '90')

=item I<opt_w>
plot width in pixels (Default: 800)

=item I<opt_h>
plot height in pixels (Default: 500)

=item I<opt_v>
verbose switch (Default: 0)

=item I<fuzzy>
fuzzy matching (Default: 0)

=item I<af>

address family, either AF_INET(default) or AF_INET6

=item I<force>

Create a plot, even when there is a cached one

=back

=head3 IMPLEMENTATION NOTES

This plots averages of averages. A lot of values need to be considered.
Consequently it takes long!
It also means that the value is undefined if one average is undefined, 
e.g. data from just one probe is missing; hence maxpct needs to be low!

=head3 RETURN VALUES

return value is string with first letter coding the result
see plot_server for explanations

=head3 SEE ALSO

see L</plot_server> for explanations

=cut

sub plot_domain {
    my $startTime=[gettimeofday];
    my $func = 'DNSRRD::plot_domain:';

    # argument processing
    my %a = (%DEFS_DOM_PLOT, @_); # argment hash values overwrite defaults
    if ($opt_d) {
        foreach my $k (sort keys %a) {
            $log->debug("$k = $a{$k}");
        }
    }

    # we need at least these
    return("E $func not 'start' argument") unless $a{start};
    return("E $func not 'stop' argument") unless $a{stop};
    return("E $func not 'domain' argument") unless $a{domain};

    # localise arguments for convenience
    my $cf = $a{cf};
    my $domain = $a{domain};
    my $maxpct = $a{maxpct};
    my $minmons = $a{minmons};
    my $opt_h = $a{opt_h};
    my $opt_l = $a{opt_l};
    my $opt_t = $a{opt_t};
    my $opt_v = $a{opt_v};
    my $fuzzy = $a{fuzzy};
    my $opt_w = $a{opt_w};
    my @orange = split(/:/, $a{orange});
    my $red = $a{red};
    my $start = $a{start};
    my $stop = $a{stop};
    my $valid = $a{valid};
	my $af    = $a{af};
	my $rrddir = $af == &AF_INET6 ? $RRDDIR6 : $RRDDIR;
	my $ipv = $af == &AF_INET ? "IPv4" : $af == &AF_INET6 ? "IPv6" : "Unknown";

    # arg checks
    $opt_t = 1 unless $opt_t;
    $cf = 'AVERAGE' unless $cf;
    $maxpct = 50 unless $maxpct;
    $maxpct = 100 if $maxpct > 100;
    my $scale = 1 / $maxpct;
    my $miny = -0.1;
    my $starts = $start->strftime($RRDTOOL_AT_TIME_FMT);
    my $stops = $stop->strftime($RRDTOOL_AT_TIME_FMT);

    my $file = mk_pngname('D', $start, $stop, $domain, %a);

	if (not exists($a{refresh})) {
		return("C $PNGPATH/$file") if (-r "$PNGDIR/$file"); # cached
	} else {
		unlink("$PNGDIR/$file");
		$log->info("Refreshing $PNGDIR/$file");
	}
    

    if ($fuzzy) {
        my $offset = 0;

        if (($stop-$start) <= 48*60*60) {
            $offset = 60*60;
        }
        else {
            $offset = 24*60*60;
        }

        my $fstart = $start - $offset;
        my $fstop = $stop - $offset;
        my $fuzzyfile = mk_pngname('D', $fstart, $fstop, $domain, %a);

	$log->warn("fuzzy mismatch on $PNGPATH/$fuzzyfile, file not found!") if (! -r "$PNGDIR/$fuzzyfile");
        return("P $PNGPATH/$fuzzyfile") if (-r "$PNGDIR/$fuzzyfile"); # fuzzyily cached
    }

	mk_pngpath($file);
    $locker->lock("$PNGDIR/$file") || $log->logdie("$func: cannot lock $PNGDIR/$file");

    if (-r "$PNGDIR/$file") { # somebody else made it while we were waiting ...
        $log->debug("$file was generated while we slept delta=",tv_interval($startTime));
        $locker->unlock("$PNGDIR/$file") || $log->logdie("$func: cannot unlock $PNGDIR/$file");
        return("C $PNGPATH/$file");
    }
    $log->debug("Generating new image: $file");
    my @servers = split(/[ \t\n]+/, $dns_config->get_domainkeys($domain,$af));
    
    @servers = grep($_=ser_alias($_), @servers);
    foreach my $s (@servers) {
        if (known_server($s,$af) < $MINPROBEVAL) {
            $locker->unlock("$PNGDIR/$file") || $log->logdie("$func: cannot unlock $PNGDIR/$file");
            return("E Server '$s' configured for domain '$domain' has insufficient $ipv data.") 
        }
    }
        
    my @mons = sort (select_mons_domain($starts, $stops, $valid, $domain, $af, &RIPE::DNSMon::RRD::all_mons($af)));
    if ($#mons <= $minmons) {
	$log->warn("$#mons <= $minmons select_mons_domain($starts, $stops, $valid, $domain, $af, &RIPE::DNSMon::RRD::all_mons($af))");
        my $msg = "<h2>";
        $msg .= $#mons<0? "No" : sprintf "Only %d", $#mons+1;
        $msg .= " probes with sufficient valid $ipv data <br>";
        $msg .= "for domain $domain<br>"; 
        $msg .= "and period $starts to $stops</h2>";
        symlink($NOGRAPH, "$PNGDIR/$file") || $log->logdie("$func: cannot symlink($NOGRAPH, $PNGDIR/$file)");
        $locker->unlock("$PNGDIR/$file") || $log->logdie("$func: cannot unlock $PNGDIR/$file");
        return("M $msg");
    }

    
    # make short names
    my %snam = &mknames(@servers);
    my %mnam = &mknames(@mons);
    
    # general plot options
    my(@go) = ();   # RRD::graph options
    push @go, "$PNGDIR/$file";
    push @go, '-a', 'PNG';
    push @go, '-s', $starts;
    push @go, '-e', $stops;
    push @go, '-w',  $opt_w ? "$opt_w" : '800';
    push @go, '-h',  $opt_h ? "$opt_h" : '500';
    push @go, '-l', sprintf '%f', $miny;
    push @go, '--rigid';
    push @go, '--title', sprintf "$ipv Unanswered Queries for Domain '$domain' from $#mons Probes ($cf)  [$starts - $stops UTC]", $domain;
    push @go, '-v', "$cf of Average Unanswered Queries (%) [0-$maxpct]";
    push @go, '-y', '1:1';
    
    # define variables from RRD
    foreach my $s (@servers)  {
        foreach my $mon (@mons) {
            next unless -f "$rrddir/$mon/$s";
            push @go, "DEF:$snam{$s}_$mnam{$mon}_D=$rrddir/$mon/$s:DROPS:$cf";
            push @go, "DEF:$snam{$s}_$mnam{$mon}_M=$rrddir/$mon/$s:DROPS:MAX";
    
        } 
    }
    
    # now define variables to plot
    my $stack = 0;
    foreach my $s (@servers)  {
        # collect average of drops
        my($cmd) = "CDEF:$snam{$s}_AD=0";
        foreach my $mon (@mons) {
            $cmd .= ",$snam{$s}_$mnam{$mon}_D,+";
        } 
        $cmd .= sprintf ',%d,/', $#mons;
        push @go, $cmd;
    
        # collect average of MAX(drops)
        $cmd = "CDEF:$snam{$s}_AM=0";
        foreach my $mon (@mons) {
            $cmd .= ",$snam{$s}_$mnam{$mon}_M,+";
        } 
        $cmd .= sprintf ',%d,/', $#mons;
        push @go, $cmd;
    
        # L - scale down to [0:1]
        push @go, "CDEF:$snam{$s}_L=$snam{$s}_AD,$scale,*";
        # C - cut off at 1 (paranoid)
        push @go, "CDEF:$snam{$s}_C=$snam{$s}_L,1,GT,1,$snam{$s}_L,IF"; 
        # G(reen)
        push @go, "CDEF:$snam{$s}_G=$snam{$s}_AM,66,GT,UNKN,$snam{$s}_C,IF";
        # Y(ellow)
        push @go, "CDEF:$snam{$s}_Y=$snam{$s}_AM,66,GT,$snam{$s}_AM,90,LT,+,2,EQ,$snam{$s}_C,UNKN,IF";
        # R(ed)
        push @go, "CDEF:$snam{$s}_R=$snam{$s}_AM,90,LT,UNKN,$snam{$s}_C,IF";
        # W(hite)
        push @go, "CDEF:$snam{$s}_W=1,1,$snam{$s}_C,-,MIN";
        # B(lack)
        push @go, "CDEF:$snam{$s}_B=$snam{$s}_AD,UN,1,0,IF";
    
        # use the pens
		# XXX  FIXME stack usage not compatible with rrdtool 1.2.x!
        if ($stack == 0) {
            push @go, sprintf "AREA:$snam{$s}_G#$YELLOW:%02d %-20.20s", $stack, &ser_name($s);
            push @go, "STACK:$snam{$s}_Y#FF8000:";
            push @go, "STACK:$snam{$s}_R#$RED:";
            push @go, "STACK:$snam{$s}_W#$GREEN:";
            push @go, "STACK:$snam{$s}_B#$WHITE:";
        }
        else {
            push @go, sprintf "STACK:$snam{$s}_G#$YELLOW:%02d %-20.20s", $stack, &ser_name($s);
            push @go, "STACK:$snam{$s}_Y#FF8000:";
            push @go, "STACK:$snam{$s}_R#$RED:";
            push @go, "STACK:$snam{$s}_W#$GREEN:";
            push @go, "STACK:$snam{$s}_B#$WHITE:";
        }
        $stack++;
    }
    
    while ($stack%5) {
           push @go, "HRULE:$miny#$WHITE:                       ";
           $stack++;
    }
    # print legend 
    push @go, "HRULE:$miny#FF8000:66% < AVG(Max Unanswered Queries) < 90%     ";
    push @go, "HRULE:$miny#$RED:90% <= AVG(Max Unanswered Queries)     ";
    push @go, "HRULE:$miny#$WHITE:no data available  ";
    
    # finally render things
    $log->debug("graph(@go)") if ($log->is_debug && $opt_d);
    my($gret, $x, $y) = RRDs::graph @go;
    if (my $ERROR = RRDs::error) {
		$log->error("RRDs::error returned: \"$ERROR\"");
        $locker->unlock("$PNGDIR/$file") || $log->logdie("$func: cannot unlock $PNGDIR/$file");
        return("E $func: RRDs::graph ERROR: $ERROR\n");
    }
    $log->debug("$file generated, delta=",tv_interval($startTime));
    
	# XXX ????? summary.png ???
    $log->debug($opt_l ? "$PNGDIR/summary-${opt_l}.png" : "$PNGDIR/summary.png") if $opt_v; 
    $log->debug(": $x x $y ", (join "/", @$gret)) if $opt_v; 

    chmod(0664, "$PNGDIR/$file") || $log->logdie("chmod(0664, $PNGDIR/$file)");
    $locker->unlock("$PNGDIR/$file") || $log->logdie("$func: cannot unlock $PNGDIR/$file");
    return("F $PNGPATH/$file");
}


# cache of last update times
# index is <probe>%<server>%<af>
my %lasttime =();

=item rrdlastt

=head3 NAME

rrdlastt -- return time-stamp of last value for this pair

=head3 SYNOPSIS

C<rrdlastt(I<$probe>, I<$server>,I<$af>)>

=head3 DESCRIPTION

rrdlastt returns the time stamp of the last value entered for the I<$probe> and
I<$server> pair, for the given address family (I<$af>, either the default
AF_INET or AF_INET6)

side effect is to create the rrd if it does not exist

=head3 RETURN VALUES

=over

=item I<$time_stamp>

Time stamp in seconds since the epoch.

=back

=cut

sub rrdlastt {
    my($probe, $server,$af) = @_;
	my $rrddir = $af == &AF_INET6 ? $RRDDIR6 : $RRDDIR;
    my($dir) = "$rrddir/$probe";
    my($rrd) = "$dir/$server";
    my($last);

    return ($lasttime{"$probe%$server%$af"}) if $lasttime{"$probe%$server%$af"};
    unless (-d $dir) {
        mkdir($dir, 0775) || $log->error("mkdir($dir, 0775)");
        chmod(0775, $dir) || $log->error("chmod(0775, $dir)");
    }
    &rrdcreate($probe, $server,$af) unless (-f $rrd);

    $last = RRDs::last $rrd;
    $log->debug("last($rrd)") if $opt_d;
    if (my $ERROR = RRDs::error) {
        $log->logdie("$0: unable to get last($rrd): $ERROR");
    }
    $lasttime{"$probe%$server%$af"} = $last;
    return($last);
}


=item rrdupdate

=head3 NAME

rrdupdate -- add a data point

=head3 SYNOPSIS

C<rrdupdate(I<$af>,I<$probe>, I<$time>, I<$server>, I<$rtt>, I<$tries>)>


=head3 DESCRIPTION

Updates the RTT and DROPS datasources. The right RRD is selected by the following parameters:

=over

=item I<$af>

Address family this packet used. Either AF_INET or AF_INET6

=item I<$probe>

Name of the originating TTM probe

=item I<$server>

Name of the DNS nameserver being monitored through the probe

=back

The following is being used in the update

=over

=item I<$time>

Time of packet being send, in seconds since the epoch

=item I<$rtt>

round trip time measured

=item I<$tries>

retries seen. on 3 retries the packet is considered lost and is registered as a drop.

=back

=head3 RETURN VALUES

=over

=item I<update_valid>

true when the time stamp of the packet is newer than the last measured value in
the RRD. However, any RRD update failures are silently ignored and logged using Log::Log4perl

=back

=cut

# New function in order to prevent ABI/API breakage.

sub rrdupdate {
    my($af,$probe, $time, $server, $rtt, $tries) = @_;
	my $rrddir = $af == &AF_INET6 ? $RRDDIR6 : $RRDDIR;
    $server =~ tr/[A-Z]/[a-z]/;
    my($rrd) = "$rrddir/$probe/$server";

    unless ($time <= &rrdlastt($probe, $server,$af)) {

        if ($tries == 1) {
            RRDs::update $rrd, "$time:$rtt:0";
        }
        elsif ($tries == 2) {
            RRDs::update $rrd, "$time:$rtt:50";
        }
        elsif ($tries == 3) {
            RRDs::update $rrd, "$time:U:100";
        }
        else {
            $log->logdie("unexpected tries value: $_");
        }

        if (my $ERROR = RRDs::error) {
                $log->error("$0: unable to update $rrd: $ERROR\n");
        }

        $lasttime{"$probe%$server%$af"} = $time;
        return(1);
    }
    return(0);
}

=item rrdcreate

=head3 NAME

rrdcreate -- create rrd archive

=head3 SYNOPSIS

C<rrdcreate(I<$probe>, I<$server>,I<$af>)>

=head3 DESCRIPTION

This function creates a new RRD file for the given I<$probe>, I<$server> and I<$af>.

=over

=item I<$probe>

Name of the TTM probe used to gather data

=item I<$server>

Name of the DNS server that is being monitored

=item I<$af>

Address family this RRD covers, either AF_INET (default) or AF_INET6

=back

The RRD has the following two data sources (DS) defined:

=over

=item RTT

The RTT datasource is a GAUGE and the heartbeat is controlled by the global
constant I<$HEARTBEAT>. The minimum allowed value is '0', there is no limit on
the maximum value.

The RTT is interpreted as milliseconds

=item DROPS

The DROPS datasource is a GAUGE and the heartbeat is controlled by the global
constant I<$HEARTBEAT>. The minimum allowed value is '0', it's maximum is '100'

The DROPS is interpreted as the percentage of packets send, which explains the maximum of '100'

=back

Each of these data sources have a set of round robin archives associated,
They are defined in the global variable I<@rras>:

=over

=item 30 days of 1/min

=item 100 days of 1/5min

=item 400 days of 1/h

=back

Each RRA uses the following consolidation functions, 
They are defined in the global variable I<@cfs>:

=over

=item AVERAGE

=item MIN

=item MAX

=back

Other global paramters controlling RRD file creation are:

=over

=item I<$STARTT>

The archive start time. This has been set to 31.7.03, the day that DNSMon started

=item I<$STEP>

The RRD stepping time is 60 seconds

=back

=cut

sub rrdcreate {

    my($probe, $server,$af) = @_;
	my $rrddir = $af == &AF_INET6 ? $RRDDIR6 : $RRDDIR;
	my($rrd) = "$rrddir/$probe/$server";
    my(@options) = ();
    my($v, $r, $c) = '';


    unless (-d "$rrddir/$probe") {
        mkdir("$rrddir/$probe", 0770) or $log->logdie("mkdir($rrddir/$probe): $!");
    }

    @options = ('-b', $STARTT, '-s', $STEP);

    push @options, "DS:RTT:GAUGE:$HEARTBEAT:0:U";
    foreach $r (@rras) {
        foreach $c (@cfs) {
            push @options, "RRA:$c:$r";
        }
    }
    push @options, "DS:DROPS:GAUGE:$HEARTBEAT:0:100";
    foreach $r (@rras) {
        foreach $c (@cfs) {
            push @options, "RRA:$c:$r";
        }
    }
    $log->debug("create($rrd, @options)") if $opt_d;
    RRDs::create $rrd, @options;
    if (my $ERROR = RRDs::error) {
          $log->logdie("$0: unable to create $rrd: $ERROR");
    }
    $lasttime{"$probe%$server%$af"} = $STARTT;
}


=item probe_last_rrd

=head3 NAME

probe_last_rrd -- return time of last data point recorded for this probe

=head3 SYNOPSIS

C<probe_last_rrd(I<$probe>,I<$af>)>

=head3 DESCRIPTION

Check the I<$probe> what the time of the last update for the RRD file was, for the given address family (I<$af>).
Currently I<$af> can be either AF_INET or AF_INET6. These constants are loaded
from Socket and Socket6. When unspecified defaults to AF_INET.

=head3 RETURN VALUES

=over

=item I<$last_update>

time of last update seen in seconds since the epoch for the given address family in the RRD file

=back


=cut

sub probe_last_rrd {
    my($probe,$af) = @_;
	my $rrddir = $af == &AF_INET6 ? $RRDDIR6 : $RRDDIR;
    my($dir) = "$rrddir/$probe";
    my($latest) = 0;

    foreach my $rrd (<$dir/*>) {
        my $last = RRDs::last $rrd;
        $log->debug("last($rrd) = $last") if $opt_d;
        if (my $ERROR = RRDs::error) {
            $log->logdie("$0: unable to get last($rrd): $ERROR");
        }
        $latest = $last if ($latest < $last);
    }
    return($latest);
	$log->debug("latest=$latest") if $opt_d;
}

=item file_last_data

last data point in this raw data file

=cut

sub file_last_data {
    my($file) = @_;
    my($latest) = 0;

# tt55.ripe.net 1050969955 ns-uk.nic.at 0.059627 1 hostname.bind ns-uk
    
    my $cmd = "tail -5 $file";
    open(LAST, "$cmd |") || $log->logdie("file_last_data($file): execute($cmd)");
    while(<LAST>) {
        if (/^([^ ]+) ([0-9]+) ([^ ]+) ([0-9.]+) ([1-3])/) {
            $latest = $2 if $2 > $latest;
        }
    }
    close(LAST);
    return($latest);
}


=item file_first_data

first data point in this raw data file

=cut

sub file_first_data {
    my($file) = @_;
    my($first) = 0;

# tt55.ripe.net 1050969955 ns-uk.nic.at 0.059627 1 hostname.bind ns-uk
    
    my $n = 0;
    open(FIRST, "<$file") || $log->logdie("file_first_data($file): open($file)");
    while(<FIRST>) {
        if (/^([^ ]+) ([0-9]+) ([^ ]+) ([0-9.]+) ([1-3])/) {
            $first = $2 if ($first == 0 || $2 < $first);
        }
        last if $n++ > 5;
    }
    close(FIRST);
    return($first);
}


=item probe_files_todo

files still to add for probe

=cut

sub probe_files_todo {
    my($probe,$af) = @_;
    my($dir) = "$DATADIR";
    my(@files) = ();

	my $fprefix = $af == &AF_INET6 ? 'DNS6' : 'DNS';
    my $lastrrd  = probe_last_rrd($probe,$af);

    foreach my $file (reverse sort <$dir/$fprefix.$probe.*>) {
        last if file_last_data($file,$af) <= $lastrrd;
        push(@files, $file);
    }
    return(sort @files);
}


=item inst_rep

report server instance used

Under http://dnsmon.example.net/dns-servmon/server/plot, select a plot on demand. manually replace type=drops with  type=instance

=cut

sub inst_rep {

    my $func = 'inst_rep';
    # argument processing
    my %a = @_;
    # we need at least these
    return("E $func no 'start' argument") unless $a{start};
    return("E $func no 'stop' argument") unless $a{stop};
    return("E $func no 'srv' argument") unless $a{srv};
    return("E $func no 'probe' argument") unless $a{probe};
    #
    my $aprobe = $a{probe};
    my $aserver = $a{srv};
    my $afrom = $a{start};
    my $ato = $a{stop};
	my $af = $a{af};

    my $lastinst = '';
    my $ret = '';

    my @todo = probe_files($aprobe, $afrom, $ato,$af);
    if ($#todo < 0) {
        $ret = "$aprobe: data not available for period requested.\n";
    }
    else {
######         tt55.ripe.net 1050969955 ns-uk.nic.at 0.059627 1 hostname.bind ns-uk

        @todo = sort @todo;
    
        foreach my $f (@todo)  {
            open(IN, "<$f") || $log->logdie("open($f)");
            while (<IN>) {
                if (/^([^ ]+) ([0-9]+) ([^ ]+) ([0-9.]+) ([1-3]) ([^ ]+) ([^ ]+)/) {
                    my($probe, $time, $server, $rtt, $try, $type, $value) = ($1, $2, $3, $4, $5, $6, $7);
                    next if $type ne 'hostname.bind' && $type ne 'id.server';
                    next if $value eq "UNREACH\n";
                    next if $server ne $aserver;
                    next if $time < $afrom->epoch;
                    last if $time > $ato->epoch;
                    if ($value =~ /jns[0-9]-([a-z]+)/) {
                        $value = "jns*-$1\n";
                    }
                    if ($lastinst ne $value) {
                        $lastinst = $value;
                        $ret .= sprintf "%-20.20s %20.20s %s", $probe, Time::Piece->strptime($time, '%s'), $value;
                    }
                }
                else {
                    $log->error("data file syntax: $_");
                }
            }
            close(IN);
        }
    }
    return("$ret\n");
}


=item probe_files

=head3 NAME

probe_files -- probe raw data files for data (from, to)

=head3 SYNOPSIS

C<probe_files(I<$probe>, I<$from>, I<$to>, I<$af>)>

=head3 DESCRIPTION

Selects usable files for I<$probe> between I<$start> and I<$stop> times  for
the given address family

=over

=item I<$probe>

Probe where we want to see the files from

=item I<$from>

Start of period, time as a Time::Piece object.

=item I<$to>

End of period, time as a Time::Piece object.

=item I<$af>

Address family, either AF_INET (default) or AF_INET6

=back


=head3 RETURN VALUES

=over

=item I<@raw_files>

Sorted list of raw files fitting in the requested period.

=back

=item NOTES

Exclusively used for the "report instance hack

=cut

sub probe_files {
    my($probe, $from, $to, $af) = @_;
    my($dir) = "$DATADIR";
    my(@files) = ();
	my $fprefix = $af == &AF_INET6 ? 'DNS6' : 'DNS';

    $log->logdie("probe_files($probe, $from, $to) from>=to") if $from>=$to;

    foreach my $file (reverse sort <$dir/${fprefix}.$probe.*>) {
        last if file_last_data($file) <= $from->epoch;
        push(@files, $file) unless file_first_data($file) > $to->epoch;
    }
    return(sort @files);
}


=item probe_last_data

=head3 NAME

probe_last_data -- last data point availabe for this probe 

=head3 SYNOPSIS

C<probe_last_data(I<$probe>,I<$af>)>

=head3 DESCRIPTION

Check the I<$probe> what the time of the last update for the raw data file was, for the given address family (I<$af>).
Currently I<$af> can be either AF_INET or AF_INET6. These constants are loaded
from Socket and Socket6. When unspecified defaults to AF_INET.

=head3 RETURN VALUES

=over

=item I<$last_update>

time of last update seen in seconds since the epoch for the given address family in the raw data file

=back

=cut

sub probe_last_data {
    my($probe,$af) = @_;
    my($latest) = 0;

	my $fprefix = $af == &AF_INET6 ? 'DNS6' : 'DNS';

    # tt55.ripe.net 1050969955 ns-uk.nic.at 0.059627 1 hostname.bind ns-uk
    
    opendir(DIR, $DATADIR) || $log->logdie ("opendir: $DATADIR");
        my @files = grep(/^${fprefix}\.$probe\./, readdir(DIR));
    close(DIR);
    return(0) if $#files < 0;
    @files = sort @files;
    
    my $cmd = "tail -5 $DATADIR/$files[$#files]";
    open(LAST, "$cmd |") || $log->logdie("last_data($probe): execute($cmd)");
    while(<LAST>) {
        if (/^([^ ]+) ([0-9]+) ([^ ]+) ([0-9.]+) ([1-3])/) {
            $latest = $2 if ($2 > $latest);
        }
    }
    close(LAST);
    return($latest);
}


=item do_updates

=head3 NAME

do_updates -- process update files

=head3 SYNOPSIS

C<do_updates(I<@files>,I<$af>)>

=head3 DESCRIPTION

Process the list of files in I<@files> for the given address family (I<$af>,
either AF_INET of AF_INET6 and defaults to AF_INET)

This will update the RRD files for any nameserver measured per probe 

=head3 BUGS

rrd files are updated per value. This will result in many open/close's of RRD
files and per update aggregation which will rewrite the file for a large part.
Batching updates would be a possible improvement.

=cut

sub do_updates(\@$) {
    my ($files_ref,$af) = @_;

    $log->logdie("\@files is not a reference") unless ref($files_ref);

    my @files = sort @$files_ref;

    # tt55.ripe.net 1050969955 ns-uk.nic.at 0.059627 1 hostname.bind ns-uk

    foreach my $f (@files)  {
        my $nrecs=0;
        my $nupdates=0;
        my $startt = [gettimeofday];
		my $offset;
		my $pid;

		# We want to read sorted data but
		# a) Don't want to start all over, if we can prevent it. so therefor we record the offset in the file and cache that
		# b) We don't want to use perl's sort routine. /bin/sort run in circles
		# around it. This is not because of the overhead of the sort routine
		# itself, but is due to split which uses very costly regular
		# expressions

		# We can go around it by using techniques discussed in http://www.sysarch.com/Perl/sort_paper.html (future ?)
		# or do it the proven way, by using the command line sort


		# set up read (-|) pipe.
		$pid = open(SORT_PIPE, "-|");

		# Do we have a pid defined ?
		if (defined($pid)) {

			if(!$pid) {
				# We are the child
				# Open the raw data file
				open(IN, $f) || $log->logdie("Couldn't open $f for input: $!");

				# Figure out where we had left off
				$offset = $offset_cache->get($f);

				# if undef, start from scratch
				$offset ||= 0;

				$log->debug("Opening $f, starting at offset $offset");

				# Start at that offset
				seek(IN,$offset,SEEK_SET);

				# /bin/sort is faster, so prepare to feed it some data ("|-")

				if (open(PIPE_CMD, "|-",qw(/bin/sort -s +2 -3) )) {
					# Copy data to sort process
					while(defined(my $input = <IN>)) {

						# sort process writes to STDOUT which happens to be PIPE_CMD for the parent.
						print PIPE_CMD $input;
					}
					close(PIPE_CMD);

					# No I/O errors ? store offset.
					if (not $!) {
						$offset = tell(IN);

						$offset_cache->set($f,$offset);
						$log->debug("Storing offset $offset for $f");
					} else {
						$log->info("Not storing offset $offset for $f, IO Error $! (retry on next interval)");
					}
					close(IN);
				}
				# We're completely done, exit

				exit(0);
			} else {
				# We're the parent. Lets read sorted data while it lasts
				while (defined(my $input = <SORT_PIPE>)) {
					chomp($input);
					$nrecs++;
					if ($input =~ /^([^ ]+) ([0-9]+) ([^ ]+) ([0-9.]+) ([1-3])/) {
					   $nupdates += &rrdupdate($af,$1, $2, $3, $4, $5);
					}
					else {
						$log->error("data file syntax: $input");
					}
				}
				# And properly close the sort pipe, unless it had died in which case we log an error
				# It is assumed that on error reading is aborted (by returning
				# undef and setting undef) and data has not been updated. It
				# will remain there for a retry next time
				close(SORT_PIPE) || $log->error("error while sorting: $? $!");
			}
		}
        $log->info("$f: $nrecs records, $nupdates updates,", tv_interval($startt), " seconds");
    }
}

=item clean_png

remove all PNGs from the cache with data between t1 and t2

=cut

sub clean_png {        
    my($t1, $t2) = @_;
    my($tf1, $tf2);
    my(@f) = ();
    
    return unless $t1 != $t2;

	if ($opt_flatpngdir) {
		opendir(DIR, $PNGDIR) || $log->logdie ("opendir: $PNGDIR");
		my @files = grep(/.png$/, readdir(DIR));
		close(DIR);

		# D_at_01.06.200300:00_12.06.200313:59_AVERAGE_250_0.5_800_500_.png

		foreach my $f (@files) {
			@f = split(/_/, $f);
			$tf1 = Time::Piece->strptime($f[1], '%Y%m%d%H%M');
			$tf2 = Time::Piece->strptime($f[2], '%Y%m%d%H%M');
			# before interval
			next if $t1->epoch < $tf1->epoch && $t2->epoch < $tf1->epoch; 
			# after interval
			next if $t1->epoch > $tf2->epoch && $t2->epoch > $tf2->epoch; 
			$log->debug("cleaning $f") if $opt_v;
			unlink("$PNGDIR/$f");
		}
	} else {
		# XXX need this as a local sub, so we get $t1, $t2 with it
		File::Find::find({wanted => sub {
			my ($dev,$ino,$mode,$nlink,$uid,$gid);

			if ( /^.*\.png\z/s && (($dev,$ino,$mode,$nlink,$uid,$gid) = lstat($_)) && -f _ ) {

				# $PNGDIR/SD/200701171246/200701181246/sec3.apnic.net/AVERAGE_250_100_800_50.png
				# $PNGDIR/D_200701152200_200701162159_com_AVERAGE_50_1_800_500.png
				
				# Extract part after ${PNGDIR} so we can handle both legacy
				# flat files that might be still left and the new directory
				# structured ones
				my ($f) = $File::Find::name =~ /^\Q${PNGDIR}\E\/?([\w_\-\/\.]+)$/;
				if (defined($f)) {
					my @f = split(/[_\/]/, $f); 
					my $tf1 = Time::Piece->strptime($f[1], '%Y%m%d%H%M');
					my $tf2 = Time::Piece->strptime($f[2], '%Y%m%d%H%M');
					# before interval
					next if $t1->epoch < $tf1->epoch && $t2->epoch < $tf1->epoch; 
					# after interval
					next if $t1->epoch > $tf2->epoch && $t2->epoch > $tf2->epoch; 
					
					$log->debug("cleaning $f");
					unlink("$PNGDIR/$f") || $log->warn("Unable to remove $PNGDIR/$f ($!)");
				} else {
					$log->warn("Not a DNSMon png file: ${File::Find::name}");
				}
				$log->debug($File::Find::name) if $opt_d;
			}
		}}, $PNGDIR);
	}
}

=item server_addr

=head3 NAME

server_addr -- lookup address from dnsmon config

=head3 SYNOPSIS

C<server_addr(I<$name>,I<$af>)>

=head3 DESCRIPTION

server_addr address uses I<addr> and I<addr6> lines in the DNSMon configuration
to lookup addresses without using resolvers. This allow us to find out
addresses that might not be configured in the DNS. Applicable use is for
deep linking to looking glasses et al.

=over

=item I<$name>

name to lookup

=item I<$af>

Return address for this address family, either AF_INET (default) or AF_INET6

=back

=head3 RETURN VALUES

=over

=item I<$addr>

The address of the server, or undef if not found.

=back

=cut

sub server_addr {
    my ($server,$af) = @_;
    return $dns_config->get_addrs($af)->{$server};
}


1;

=back

=head1 AUTHOR 

Information services, RIPE NCC <dnsmon@ripe.net>

=cut
