package RIPE::DNSMon::TestMB;
############################################################################
###    (C)opyright 2003 - 2008 RIPE NCC
###    This file is part of DNSMon
###
###    DNSMon is free software: you can redistribute it and/or modify
###    it under the terms of the GNU General Public License as published by
###    the Free Software Foundation, either version 3 of the License, or
###    (at your option) any later version.
###
###    DNSMon is distributed in the hope that it will be useful,
###    but WITHOUT ANY WARRANTY; without even the implied warranty of
###    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
###    GNU General Public License for more details.
###
###    You should have received a copy of the GNU General Public License
###    along with DNSMon.  If not, see <http://www.gnu.org/licenses/>.
############################################################################

use strict;
use warnings;
use base 'RIPE::ApacheMysql::TestMB';
use Date::Simple;

our $VERSION='0.01';

sub new {
    my $pkg = shift;
    my $self = $pkg->SUPER::new(@_);

    $self->{properties}{dnsmon_test_args} ||= {};
    my $p = $self->{properties}{dnsmon_test_args};

    $p->{test_dir}  ||= $self->base_dir."/_dnsmon";
    $p->{rrddir}    ||= $p->{test_dir}."/rrd";
    $p->{rrddir6}    ||= $p->{test_dir}."/rrd6";
    # XXX can Apache::TestMB tell me the DocumentRoot?
    $p->{pngdir}    ||= $self->base_dir."/t/htdocs/png";
    $p->{datadir}   ||= $p->{test_dir}."/datadir";
    $p->{pngpath}   ||= "/png";
#    $p->{nograph}   ||= $self->base_dir."/t/htdocs/dnsmon/shared/no-graph.png";
    $p->{config}    ||= $self->base_dir."/dnsmon.conf";

    $p->{pkg_conf_file}    ||= $p->{test_dir}."/dnsmon-rrd.conf";
    $p->{pkg_conf_in_file} ||= $self->base_dir."/dnsmon-rrd.conf.in";

    $self->add_to_cleanup('_mason');
    return $self;
}

sub find_etc_files {
   my $self = shift;
   my %files;
   File::Find::find( sub {
     return unless -f && /\.conf$/;
     $files{$File::Find::name} = $File::Find::name;
  }, "etc" );
  return \%files;
}

sub valid_property {
    return 1 if defined $_[1] && (
        $_[1] eq 'dnsmon_test_args'
    );
    shift->SUPER::valid_property(@_);
}

sub ACTION_test {
    my $self = shift;

    $self->depends_on('dnsmon_conf');
    $self->SUPER::ACTION_test(@_);
}

sub ACTION_start_daemons {
    my $self = shift;

    $self->depends_on('dnsmon_conf');
    $self->SUPER::ACTION_start_daemons(@_);
}

sub ACTION_dnsmon_conf {
    my $self = shift;

    my $p = $self->{properties}{dnsmon_test_args};

    if (-d $p->{test_dir}) {
        # clean up from previous run
        $self->delete_filetree($p->{test_dir});
    }
    File::Path::mkpath($p->{test_dir});
    File::Path::mkpath($p->{pngdir});
    File::Path::mkpath($p->{datadir});
    File::Path::mkpath($p->{rrddir});
    File::Path::mkpath($p->{rrddir6});

    my $conf = $p->{pkg_conf_file};
    my $conf_in = $p->{pkg_conf_in_file};
  
    warn "DNSMon: Writing Test configuration to $conf\n";
  
    unless( -e $conf_in ) {
        die "can't find the conf input file\n";
    }
    open CIN, "$conf_in" or die "can't read input conf file $conf_in: $!";
    open COUT, ">$conf" or die "can't write conf file $conf: $!";

    my $docroot = $self->base_dir.'/t/htdocs';
  
    while (<CIN>) {
        s/\@RrdDir\@/$p->{rrddir}/g;
        s/\@RrdDir6\@/$p->{rrddir6}/g;
        s/\@PngDir\@/$p->{pngdir}/g;
        s/\@DataDir\@/$p->{datadir}/g;
        s/\@PngPath\@/$p->{pngpath}/g;
        s/\@DocumentRoot\@/$docroot/g;
        s/\@Config\@/$p->{config}/g;
        print COUT;
    }
    close CIN;
    close COUT;
 
    # critical, set the ENV so RIPE::DNSMon::RRD will pick up the test config
    $ENV{DNSMONCONFIG} = $p->{pkg_conf_file};
}

sub ACTION_htdocs_to_test_docroot {
    my $self = shift;
    my $base = $self->base_dir;
    $self->_sync('', "$base/htdocs/",    "$base/t/htdocs/" );
	symlink("dhandler","$base/t/htdocs/dns-servmon/domain/index.html");
	symlink("dhandler","$base/t/htdocs/dns-servmon/server/index.html");
	symlink("dhandler","$base/t/htdocs/dns-servmon/probe/index.html");
}

sub ACTION_htdocs_fakeinstall {
    my $self = shift;
    my $base = $self->base_dir. '/htdocs';
    my $remote = '/export/dnsmon/www';
    $self->_sync('n', "$base/favicon.ico",   "$remote/favicon.ico");
    $self->_sync('n', "$base/dnsmon.css",   "$remote/dnsmon.css");
    $self->_sync('n', "$base/index.html",   "$remote/index.html");
    $self->_sync('n', "$base/dns-servmon/", "$remote/dns-servmon/");
    $self->_sync('n', "$base/images/",      "$remote/images/");
    $self->_sync('n', "$base/js/",      "$remote/js/");
    $self->_sync('n', "$base/ripencc/",     "$remote/ripencc/");
    $self->_sync('n', "$base/robots.txt",   "$remote/robots.txt");
	print "symlink(\"dhandler\",\"$remote/dns-servmon/domain/index.html\")\n";
	print "symlink(\"dhandler\",\"$remote/dns-servmon/server/index.html\")\n";
	print "symlink(\"dhandler\",\"$remote/dns-servmon/probe/index.html\")\n";
}

sub ACTION_htdocs_install {
    my $self = shift;
    my $base = $self->base_dir. '/htdocs';
    my $remote = '/export/dnsmon/www';
    $self->_sync('', "$base/favicon.ico",   "$remote/favicon.ico");
    $self->_sync('', "$base/dnsmon.css",   "$remote/dnsmon.css");
    $self->_sync('', "$base/index.html",   "$remote/index.html");
    $self->_sync('', "$base/dns-servmon/", "$remote/dns-servmon/");
    $self->_sync('', "$base/images/",      "$remote/images/");
    $self->_sync('', "$base/js/",      "$remote/js/");
    $self->_sync('', "$base/ripencc/",     "$remote/ripencc/");
    $self->_sync('', "$base/robots.txt",   "$remote/robots.txt");
	symlink("dhandler","$remote/dns-servmon/domain/index.html");
	symlink("dhandler","$remote/dns-servmon/server/index.html");
	symlink("dhandler","$remote/dns-servmon/probe/index.html");
}

sub _sync {
    my ($self, $n, $source, $target) = @_;
	$n ||= '';

	$self->do_system("rsync -${n}Cav --no-owner --delete $source $target");
}


1;
__END__

=head1 NAME

RIPE::Travel::TestMB - apache/mysql+travel testing

=head1 SYNOPSIS 

 #### in Build.PL

 use RIPE::Travel::TestMB;

 my $build = RIPE::Travel::TestMB->new(
    module_name => 'My::Module',
 );
 $build->create_build_script;

 #### run tests with mysql server
 ./Build 
 ./Build test # starts httpd & mysql before running tests

 #### start daemons and leave running
 ./Build start_daemons

 #### stop daemon
 ./Build stop_daemons

 #### connect to test databaes with mysql util
 ./Build mysql

=head1 ACTIONS

=item htdocs_fakeinstall 

Performs a test rysnc of the files from C<apache_source_htdocs> to their live
area. Nice for seeing what will be changing.

=item htdocs_install 

Installs the files from C<apache_source_htdocs> via rsync to their live
area. 

=head1 REQUIRES

L<Module::Build>
L<RIPE::ApacheMysql::TestMB>

=head1 DESCRIPTION

A RIPE NCC hack to run httpd and mysqld for tests the travel software.
It inherits from L<RIPE::ApacheMysql::TestMB>.

Note: This module can only be used with L<Module::Build>.

=head1 TODO

Integrate into L<RIPE::ApacheMysql::TestMB>

=head1 SEE ALSO

 L<Apache::TestMB>
 L<RIPE::ApacheMysql::TestMB>
 L<Module::Build>
 http://iii.ripe.net/sed/twiki/bin/view/SED/ApacheMysqlTestMB
