/*
 *   (C)opyright 2003 - 2008 RIPE NCC
 *   This file is part of DNSMon
 *
 *   DNSMon is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   DNSMon is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with DNSMon.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <sys/types.h>

#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/udp.h>

#include "checksum.h"

static int in_cksum(u_int16_t *addr, int len);

u_int16_t
do_checksum(u_char *buf, int protocol, int len)
{
    struct ip *iph_p;
    int ip_hl;
    int sum = 0;
    struct udphdr *udph_p;

    /* locate IP header in buffer */
    iph_p = (struct ip *)buf;

    /* grab IP header length */
    ip_hl = iph_p->ip_hl << 2;

    /* locate UDP header in buffer */
    udph_p = (struct udphdr *)(buf + ip_hl);

    /* clear checksum */
    udph_p->uh_sum = 0;

    /* calculate checksum */
    sum = in_cksum((u_int16_t *)&iph_p->ip_src, 8);
    sum += ntohs(IPPROTO_UDP + len);
    sum += in_cksum((u_int16_t *)udph_p, len);

    /* get rid of carry */
    sum =  (sum >> 16) + (sum & 0xffff); 
    sum = (~(sum + (sum >> 16)) & 0xffff);

    return (sum);
}

/*
 * Actual checksum function
 */
static int
in_cksum(u_int16_t *addr, int len)
{
    int sum;
    int nleft;
    u_int16_t ans;
    u_int16_t *w;

    sum = 0;
    ans = 0;
    nleft = len;
    w = addr;

    while (nleft > 1)
    {
        sum += *w++;
        nleft -= 2;
    }
    if (nleft == 1)
    {
        *(u_char *)(&ans) = *(u_char *)w;
        sum += ans;
    }
    return (sum);
}
