/*
 *   (C)opyright 2003 - 2008 RIPE NCC
 *   This file is part of DNSMon
 *
 *   DNSMon is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   DNSMon is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with DNSMon.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <string.h>

#include "nsmd.h"

extern struct globaltime gt;

/* #define CONFIG_DEBUG */

int
read_config(struct server list[])
{
    FILE *cfd;
    char * buf = malloc(80);
    int n = 0;
    struct timeval tv;
    struct stat configstats ;
    int i;
    char configfile[100];

#ifdef CONFIG_DEBUG
    printf("%s(): read in configuration\n", __func__);
#endif /* CONFIG_DEBUG */

    strcpy (configfile, getenv("TTROOT"));
    strcat (configfile, "/config/nsmd.conf");

#ifdef CONFIG_DEBUG
    printf("configfile: %s\n", configfile);
#endif /* CONFIG_DEBUG  */

    /* check if configfile has been updated */
    if(stat (configfile, &configstats)) {
        printf("ERROR: %s(): config file not found\n", __FUNCTION__);
        exit(1);
    }

#ifdef CONFIG_DEBUG
    printf ("Configfile last read at %d\n", gt.read);
    printf ("Configfile last modified at %d\n", configstats.st_ctime);
#endif /* CONFIG_DEBUG */

  if ( gt.read > configstats.st_ctime ) {
#ifdef CONFIG_DEBUG
        printf("use old schedule\n");
#endif /* CONFIG_DEBUG */
        return (0);
    }
#ifdef CONFIG_DEBUG
    printf("create new schedule\n");
#endif /* CONFIG_DEBUG */

    /* erase list */
    for (i = 0; i < LIST_SIZE; i++) {
        list[i].active = 0;
    }

    cfd = fopen(configfile, "r");
    if (cfd < 0) {
        perror("open");
        exit(1);
    }

    /* record time of config file read */
    gettimeofday(&tv, NULL);
    gt.read = tv.tv_sec;
    
    while (fgets(buf, 80, cfd) != NULL) { /* read all lines until EOF */

        *(buf+strlen(buf)-1) = 0; /* remove \n */

        if (strlen(buf) == 0) /* empty line */
            continue;

        if (!strncmp(buf, "#", 1)) /* comments */
            continue;

        //format: server label addres query_type domain freq timeout
        if (!strncmp(buf, "server ", 7)) { /* target */
            char *bufp = buf;
            char * label;
            char * address;
            char * query_type;
            char * domain;
            char * frequency;
            char * timeout;

#ifdef IPV6_SUPPORT
			struct in6_addr in6addr;
			struct in_addr inaddr;
#endif /* IPV6_SUPPORT */

            if (n == LIST_SIZE) {
                printf("ERROR: %s(): please increase LIST_SIZE, too many targets\n", 
                        __FUNCTION__);
                exit(1);
            }

            strsep(&bufp, " \t");
            label = strsep(&bufp, " \t");
            address = strsep(&bufp, " \t");
            query_type = strsep(&bufp, " \t");
            domain = strsep(&bufp, " \t");
            frequency = strsep(&bufp, " \t");
            timeout = strsep(&bufp, " \t");


			/* 
			 * Check whether address is "supported". This prevents machines
			 * without an IPv6 API at compile time to try to send packets with
			 * a v6 destination address because they will send everything as v4
			 * by default as no v4/v6 capable address manipulation functions
			 * are available. For machines with a v6 stack, but not configured
			 * we'll save some space in the scheduler by not adding them
			 * leaving more room  for other measurements
			 */
#ifdef IPV6_SUPPORT
			/* With a V6 stack present, add v6 measurent if stack is configured. always add v4 addresses */
			if ((inet_pton(AF_INET6, address, &in6addr) && candov6) || (inet_pton(AF_INET, address, &inaddr))) {
#else
			/* On non v6 stack present boxes. skip addreses if they have a colon in it */
			if(strchr(address,':') == NULL) {
#endif

				list[n].active = 1;
				strcpy(list[n].label, label);
				strcpy(list[n].address, address);
				strcpy(list[n].domain, domain);
				list[n].query_type = atoi(query_type);
				list[n].frequency = atoi(frequency);
				list[n].timeout = atoi(timeout);

				n++; /* increase */
			}
            continue;
        }

        if (!strncmp(buf, "period ", 7)) { /* base period wrt frequency */
            char *bufp = buf;
            char * label;

            strsep(&bufp, " \t");
            label = strsep(&bufp, " \t");
            gt.period = atoi(label);

#ifdef CONFIG_DEBUG
            printf("Period: %d\n", gt.period);
#endif /* CONFIG_DEBUG */
            continue;
        }
    }
    
    fclose(cfd);

    free(buf);

    return (0);
}
