/*
 *   (C)opyright 2003 - 2008 RIPE NCC
 *   This file is part of DNSMon
 *
 *   DNSMon is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   DNSMon is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with DNSMon.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <arpa/nameser.h>
#include <netinet/udp.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>

#ifdef __FreeBSD__
#include <netinet/ip_var.h>
#endif /* __FreeBSD__ */

#include <stdio.h>
#include <stdlib.h>

#if __FreeBSD_version >= 400000
  #include <ifaddrs.h>
#else
  #include "ifaddrs.h"
#endif /* __FreeBSD_version */

#include "get_ip_address.h"

unsigned int src_ip;

/* #define INTERFACE_DEBUG */

#ifndef    INADDR_LOOPBACK 
#define INADDR_LOOPBACK (u_int32_t)0x7f000001
#endif /* INADDR_LOOPBACK */


void
get_ip_address(void)
{
    struct ifaddrs *ifap;
    struct ifaddrs *tmp;
    struct sockaddr_in * sin;

    getifaddrs(&ifap);

    for (tmp = ifap; tmp; tmp = tmp->ifa_next) {
        if (tmp->ifa_addr->sa_family == AF_INET) {
            sin = (struct sockaddr_in *) (tmp->ifa_addr);
            if (ntohl(sin->sin_addr.s_addr) == INADDR_LOOPBACK) {
#ifdef INTERFACE_DEBUG
                printf("%s: loopback\n", tmp->ifa_name);
#endif /* INTERFACE_DEBUG */
            } else {
                src_ip = ntohl(sin->sin_addr.s_addr);
                /* skip zero iface address */
                if (src_ip != 0) {
#ifdef INTERFACE_DEBUG
                    printf("%s: non loopback\n",
                        tmp->ifa_name);
                    printf("%s: picked %s\n", 
                        tmp->ifa_name,
                    inet_ntoa(sin->sin_addr));
#endif /* INTERFACE_DEBUG */
                    /* found one */
                    return;
                }
            }
                
                
        }
    }

    printf("ERROR: didn't find any usable IP address on interfaces\n");
    exit(1);
}
