/*
 *   (C)opyright 2003 - 2008 RIPE NCC
 *   This file is part of DNSMon
 *
 *   DNSMon is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   DNSMon is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with DNSMon.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <sys/types.h>
#include <sys/queue.h>
#include <sys/time.h>

#include <unistd.h>
#include <stdio.h>
#include <strings.h>
#include <string.h>
#include <stdlib.h>

#include "nsmd.h"
#include "process_data.h"
#include "receiving_process.h"
#include "sending_process.h"
#include "time.h"
#include "grimreaper.h"
#include "queue.h"

/* from TTM's libtb */
#include "utils.h"


extern SLIST_HEAD(outst_m_listhead, outst_m_entry) outst_m_head;
extern int g_pipe[2];

int check_queue(void);

/* #define GRIMREAPER_DEBUG */
 
void
grimreaper(void)
{
    check_queue();
}

int check_queue(void)
{
    struct outst_m_entry * entry, * tmp;
    struct timeval tv;
    struct timeval age_tv;
    int age;
    int queue_index = 0;
    struct grimreaper_packet gpack;
    struct merged_entry mentry;

    gettimeofday(&tv, NULL);

    SLIST_FOREACH_SAFE(entry, &outst_m_head, entries, tmp) {
        queue_index++;
		
	ncc_timersub(&tv,&entry->sent_tv,&age_tv);
#ifdef GRIMREAPER_DEBUG
        printf("index: %d age: %ld.%ld: sched_id: %d\n", 
            queue_index, age_tv.tv_sec, age_tv.tv_usec, entry->sched_id);
#endif /* GRIMREAPER_DEBUG */

        age = age_tv.tv_sec;

        if (age >= entry->timeout) {
#ifdef GRIMREAPER_DEBUG
            printf("Activating resend with sched_id: %d\n",
                entry->sched_id);
#endif /* GRIMREAPER_DEBUG */

            if (entry->attempts > 2 )
                printf("ERROR: attempts > tries\n");

            if (entry->attempts < 2) {
                bzero(&gpack, sizeof(gpack));
                gpack.sched_id = entry->sched_id;
                gpack.attempts = entry->attempts;

                if(write(g_pipe[PIPE_WRITE_END], &gpack,
                    sizeof(gpack)) != sizeof(gpack)) {
                    printf("ERROR: write to pipe failed\n");
                }

            } else {
#ifdef GRIMREAPER_DEBUG
                printf("Too many (%d) attempts for: %d - "
                    "drop\n", entry->attempts, 
                    entry->sched_id);
#endif /* GRIMREAPER_DEBUG */

                mentry.query_type = entry->query_type;
                strcpy(mentry.qry, entry->qry);
                strcpy(mentry.label, entry->label);
                strcpy(mentry.result, "UNREACH");
                strcpy(mentry.domain, entry->domain);
                mentry.attempts = 3;
                mentry.delay.tv_sec = 0;
                mentry.delay.tv_usec = 0;

                log_entry(mentry,entry->af);

            }
                
            SLIST_REMOVE(&outst_m_head, entry, 
                outst_m_entry, entries);

            free(entry);
        }
    }

    return (0);
}
