/*
 *   (C)opyright 2003 - 2008 RIPE NCC
 *   This file is part of DNSMon
 *
 *   DNSMon is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   DNSMon is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with DNSMon.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <sys/types.h>
#include <sys/select.h>
#include <sys/time.h>
#include <sys/queue.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/socket.h>
#ifdef __FreeBSD__
#include <libutil.h>
#endif /* __FreeBSD__ */
#ifdef __linux__
#include "setproctitle.h"
#endif /* __linux__ */


#include "nsmd.h"
#include "receiving_process.h"
#include "sending_process.h"
#include "process_data.h"
#include "time.h"
#include "grimreaper.h"
#include "queue.h"

/* from TTM's libtb */
#include "utils.h"


extern int r_pipe[2];
extern int s_pipe[2];

SLIST_HEAD(outst_m_listhead, outst_m_entry) outst_m_head =
    SLIST_HEAD_INITIALIZER(outst_m_head);
struct outst_m_listhead *outst_m_headp;        /* Singly-linked List head. */

char hostname[100];
FILE *output = NULL, *output6 = NULL;

char prefix[100], prefix6[100];

int handle_sent_packet(struct sent_packet spack);
int handle_received_packet(struct received_packet rpack);
int log_entry(struct merged_entry m, int af);

/* #define PROCESS_DEBUG */

int process_data()
{
    fd_set fdset;
    struct received_packet rpack;
    struct sent_packet spack;
    int ret;
    pid_t pid;

    /* create new process */
    if ((pid = fork()))
        return (pid);

    setproctitle("[process data]");

    /* get the hostname */
    gethostname(hostname, 100);

    sprintf (prefix, "%s/%s%s", getenv("TTROOT"), "/dns/DNS.", hostname);
    sprintf (prefix6, "%s/%s%s", getenv("TTROOT"), "/dns6/DNS6.", hostname);

    SLIST_INIT(&outst_m_head);

    for (;;) {
        FD_ZERO(&fdset);
        FD_SET(r_pipe[PIPE_READ_END], &fdset);
        FD_SET(s_pipe[PIPE_READ_END], &fdset);
        ret = select(max(r_pipe[PIPE_READ_END], 
            s_pipe[PIPE_READ_END]) + 1, &fdset, NULL, NULL, NULL);

        if (FD_ISSET(r_pipe[PIPE_READ_END], &fdset)) {
            if(read(r_pipe[PIPE_READ_END], &rpack, sizeof(rpack)) !=
                sizeof(rpack)) {
                printf("ERROR: read on pipe failed\n");
            } else {
                handle_received_packet(rpack);
            }
        }


        if (FD_ISSET(s_pipe[PIPE_READ_END], &fdset)) {
            if(read(s_pipe[PIPE_READ_END], &spack, sizeof(spack)) !=
                sizeof(spack)) {
                printf("ERROR: read on pipe failed\n");
            } else {
                handle_sent_packet(spack);
            }
        }
        grimreaper();
    }

	if (output)
		fclose(output);
	if (output6)
		fclose(output6);
}

int
handle_sent_packet(struct sent_packet spack)
{
    struct outst_m_entry * new;

    new = malloc(sizeof(struct outst_m_entry));
    bzero(new, sizeof(new));

    new->af = spack.af;
    new->dns_id = spack.dns_id;
    new->sched_id = spack.sched_id;
    new->sent_tv = spack.sent_tv;
    new->timeout = spack.timeout;
    new->attempts = spack.attempts;
    new->query_type = spack.query_type;
    strcpy(new->qry, spack.qry);
    strcpy(new->domain, spack.domain);
    strcpy(new->label, spack.label);

    SLIST_INSERT_HEAD(&outst_m_head, new, entries);

    return (0);
}

int
handle_received_packet(struct received_packet rpack)
{
    struct outst_m_entry * entry, * tmp;
    struct merged_entry m;

    SLIST_FOREACH_SAFE(entry, &outst_m_head, entries, tmp) {
        if(entry->dns_id == rpack.dns_id) {
            SLIST_REMOVE(&outst_m_head, entry, 
                outst_m_entry, entries);
#ifdef PROCESS_DEBUG
            printf("Removing entry with sched_id: %d\n",
                entry->sched_id);
#endif /* PROCESS_DEBUG */

            bzero(&m, sizeof(m));
            strcpy(m.label, entry->label);
			ncc_timersub(&rpack.recv_tv,&entry->sent_tv,&m.delay);
            m.attempts = entry->attempts;
            m.query_type = entry->query_type;
            strcpy(m.qry, entry->qry);
            strcpy(m.domain, entry->domain);
            strcpy(m.result, rpack.result);

            free(entry);
    
            log_entry(m,rpack.af);

            break;
        }
    }

    return (0);
}

int
log_entry(struct merged_entry m, int af)
{

    char logfile[200];
    time_t t ;
    struct tm *timevalue;
    /* set initally to zero as that identifies the first run */
    static int day = 0; 
	FILE *current_output=NULL;

    /* get unix time */
    time(&t);
    timevalue = gmtime (&t);

    /* only reopen the file if it is a new day */
    if (day != timevalue->tm_mday) {

        /* day=0 for the first run, so no need to close the file */
        if (day) {
			if (output)
				fclose(output);
			if (output6)
				fclose(output6);
		}
        /* remember the day for the next packet */
        day = timevalue->tm_mday;

        /* append date to filename prefix */
        timevalue->tm_mon++; /* tm_mon is from 0-11 */
        timevalue->tm_year += 1900;
        sprintf (logfile, "%s.%.4d%.2d%.2d", prefix, timevalue->tm_year,
            timevalue->tm_mon, timevalue->tm_mday);

        /* open output file in append mode (and create if not exist) */
        output = fopen(logfile, "a");
        if (output == NULL) {
            printf("ERROR: %s(): fopen of output file failed\n",
                __FUNCTION__);
            perror("fopen");
            exit(1);
        }

#ifdef PROCESS_DEBUG
        printf("Opening logfile: %s\n", logfile);
#endif /* PROCESS_DEBUG */

#ifdef IPV6_SUPPORT
		if(candov6) {
			sprintf (logfile, "%s.%.4d%.2d%.2d", prefix6, timevalue->tm_year,
				timevalue->tm_mon, timevalue->tm_mday);

			/* open output file in append mode (and create if not exist) */
			output6 = fopen(logfile, "a");
			if (output == NULL) {
				printf("ERROR: %s(): fopen of output file failed\n",
					__FUNCTION__);
				perror("fopen");
				exit(1);
			}
#ifdef PROCESS_DEBUG
        printf("Opening logfile: %s\n", logfile);
#endif /* PROCESS_DEBUG */
		}
#endif
    }

	current_output = af == AF_INET ? output : output6;

	if (current_output) {

		if (m.query_type == QT_SOA) {
			/* <hostname> <time> <target> <resp time> <attempt> <soa> 
			 * <domain> <version resp> */
			fprintf(current_output, "%s %ld %s %s %d %s %s %s\n",
				hostname, (long int)t, m.label, 
				timevaltoa(m.delay), 
				m.attempts, m.qry, m.domain, 
				m.result);
		} else {
			/* <hostname> <time> <target> <resp time> <attempt> <qry> 
			 * <result> */
			fprintf(current_output, "%s %ld %s %s %d %s %s\n",
				hostname, t, m.label,
				timevaltoa(m.delay), 
				m.attempts, m.qry, m.result);
		}

		/* flush out every entry to the current_output file instead of buffered write */
		fflush(current_output);
	}

    return (0);
}
