/*
 *   (C)opyright 2003 - 2008 RIPE NCC
 *   This file is part of DNSMon
 *
 *   DNSMon is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   DNSMon is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with DNSMon.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <sys/queue.h>
#include <stdio.h>
#include <stdlib.h>

#include "nsmd.h"
#include "scheduler.h"
#include "poisson.h"
#include "queue.h"

extern struct globaltime gt;

STAILQ_HEAD(, measurement) schedule; /* schedule */

/* #define SCHEDULER_DEBUG */

int
make_schedule(struct server list[])
{
    int i;
    int period = gt.period;
    long seed = -1;
    int n;
    struct measurement * tmp;

    while (!STAILQ_EMPTY(&schedule)) {
        tmp = STAILQ_FIRST(&schedule);
        STAILQ_REMOVE_HEAD(&schedule, entries);
        free(tmp);
    }
    STAILQ_INIT(&schedule);

    for (i = 0; i < LIST_SIZE && list[i].active == 1; i++) {
        int freq;
        int t;

        freq = list[i].frequency;

        /* our resolution is 1 second for now */
        if (freq > period) {
            printf("WARNING: frequency (%d) > period (%d)\n", 
                freq, period);
        }
        
        /* should be divideble */
        if (period % freq) {
            printf("WARNING: frequency (%d) %% period (%d) != 0\n",
                freq, period);
        }

        if(freq == period)
            printf("WARNING: activating 1s hack for target %d!\n", i);

        for (t = 0, n = 0; t < period && n < list[i].frequency; n++) {
            struct measurement *m = NULL;
            struct measurement *prev = NULL;
            struct measurement *next = NULL;
            if(freq != period) {
                seed = poisson_getseed(0, NULL);
                seed = do_poisson (1, list, i, seed);
                poisson_saveseed(0, NULL, seed);
            } else {
                list[i].time = 1;
            }

            t += list[i].time;
            if (t > gt.period) {
#ifdef SCHEDULER_DEBUG
                printf("this measurement time is exceeding "
                    "the period: %d\n", t);
#endif /* SCHEDULER_DEBUG */
                break;
            }
            if (n > list[i].frequency) {
#ifdef SCHEDULER_DEBUG
                printf("this measurement time is exceeding "
                    "the freq: %d\n", n);
#endif /* SCHEDULER_DEBUG */
                break;
            }

            m = malloc(sizeof(struct measurement));

            m->targetid = i;
            m->time = t;
#ifdef SCHEDULER_DEBUG    
            printf("t: %d\n", t);
#endif /* SCHEDULER_DEBUG */
            STAILQ_FOREACH(prev, &schedule, entries) {
                next = STAILQ_NEXT(prev, entries);
                if (next) {
                    /* more items coming */
                    if (next->time > t) {
                        /* we can put our entry here 
                         * because the next one is 
                         * bigger
                         */
                        break;
                    } else
                        continue;
                } else
                    break;
            }
            if (STAILQ_EMPTY(&schedule)) {
#ifdef SCHEDULER_DEBUG    
                printf("list is empty adding to tail\n");
#endif /* SCHEDULER_DEBUG */
                STAILQ_INSERT_TAIL(&schedule, m, entries);
            } else {
                if (prev->time > t) {
#ifdef SCHEDULER_DEBUG    
                    printf("adding before first\n");
#endif /* SCHEDULER_DEBUG */
                    STAILQ_INSERT_HEAD(&schedule, m,
                        entries);
                } else {
#ifdef SCHEDULER_DEBUG    
                    printf("adding after prev\n");
#endif /* SCHEDULER_DEBUG */
                    STAILQ_INSERT_AFTER(&schedule, prev,
                        m, entries);
                }
            }

        } /* for all dispersions */

    } /* for all targets */

    return (0);
}
