
/*
  Copyright (C) 2004-2006  Dmitry V. Levin <ldv@altlinux.org>

  The setproctitle library interface.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdarg.h>

#include "setproctitle.h"

extern char *program_invocation_name;
char *program_invocation_short_name;
extern char **environ;

extern int _init(int argc, char *argv[], char *envp[]);

static char *title_buffer = 0;
static size_t title_buffer_size = 0;
static char *title_progname, *title_progname_full;

int
setproctitle(const char *fmt, ...)
{
	if (!title_buffer || !title_buffer_size)
	{
		errno = ENOMEM;
		return -1;
	}

	memset(title_buffer, '\0', title_buffer_size);

	ssize_t written;

	if (fmt)
	{
		ssize_t written2;
		va_list ap;

		written =
			snprintf(title_buffer, title_buffer_size, "%s: ",
				 title_progname);
		if (written < 0 || (size_t) written >= title_buffer_size)
			return -1;

		va_start(ap, fmt);
		written2 =
			vsnprintf(title_buffer + written,
				  title_buffer_size - written, fmt, ap);
		va_end(ap);
		if (written2 < 0
		    || (size_t) written2 >= title_buffer_size - written)
			return -1;
	} else
	{
		written =
			snprintf(title_buffer, title_buffer_size, "%s",
				 title_progname);
		if (written < 0 || (size_t) written >= title_buffer_size)
			return -1;
	}

	written = strlen(title_buffer);
	memset(title_buffer + written, '\0', title_buffer_size - written);

	return 0;
}

/*
  It has to be _init function, because __attribute__((constructor))
  functions gets called without arguments.
*/
