/*
 *   (C)opyright 2003 - 2008 RIPE NCC
 *   This file is part of DNSMon
 *
 *   DNSMon is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   DNSMon is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with DNSMon.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <sys/types.h>
#include <sys/socket.h>

#include <netinet/in.h>

#include <strings.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

#include "nsmd.h"

int raw_fd;
int udp_fd;
#ifdef IPV6_SUPPORT
int udp_v6_fd;
#endif /* IPV6_SUPPORT */

/* #define SOCKET_DEBUG */
int
setup_udp_socket()
{
    struct sockaddr_in saddr;

#ifdef SOCKET_DEBUG
    printf("%s(): setup the UDP socket\n", __func__);
#endif /* SOCKET_DEBUG */

    /* open a UDP socket for network communications */
    udp_fd = socket(AF_INET, SOCK_DGRAM, 0);
    if (udp_fd < 0) {
        perror("udp socket");
        exit(1);
    }

    bzero((char *)&saddr, sizeof(saddr));
    saddr.sin_family = AF_INET;
    saddr.sin_addr.s_addr = htonl(INADDR_ANY);
    saddr.sin_port = htons(UDP_SRC_PORT);

    if (bind(udp_fd, (struct sockaddr *)&saddr, sizeof(saddr))) {
        perror("bind");
        exit(1);
    }

    return (0);
}

int
setup_raw_socket()
{
    int hincl;

#ifdef SOCKET_DEBUG
    printf("%s(): setup the raw socket\n", __func__);
#endif /* SOCKET_DEBUG */

    /* open a raw socket for network communications */
    raw_fd = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP);
    if (raw_fd < 0) {
        perror("socket");
        exit(1);
    }

    /* setup socket to expect packets including IP header */
    hincl = 1;
    setsockopt(raw_fd, IPPROTO_IP, IP_HDRINCL, &hincl, sizeof(hincl));

    return (0);
}

#ifdef IPV6_SUPPORT
int
setup_udp_v6_socket()
{
    struct sockaddr_in6 saddr;

#ifdef SOCKET_DEBUG
    printf("%s(): setup the UDP IPv6 socket\n", __func__);
#endif /* SOCKET_DEBUG */

    /* open a UDP socket for network communications */
    udp_v6_fd = socket(AF_INET6, SOCK_DGRAM, 0);
    if (udp_fd < 0) {
        perror("udp v6 socket");
        exit(1);
    }

    bzero((char *)&saddr, sizeof(saddr));
    saddr.sin6_family = AF_INET6;
    saddr.sin6_addr = in6addr_any;
    saddr.sin6_port = htons(UDP_SRC_PORT+100);

    if (bind(udp_v6_fd, (struct sockaddr *)&saddr, sizeof(saddr))) {
        perror("bind v6");
        exit(1);
    }

    return (0);
}
#endif /* IPV6_SUPPORT */

void
breakdown_sockets()
{
    close(raw_fd);
    close(udp_fd);
#ifdef IPV6_SUPPORT
    close(udp_v6_fd);
#endif /* IPV6_SUPPORT */
}
